## Filename    : edit_mvc.py
## Author(s)   : Geoffroy Andrieux
## Created     : 03/2010
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne, Nolwenn Le Meur
##
"""
Class for managing a triple (Model, View, Controler) associated with a model
"""

from cadbiom.models.guard_transitions.chart_model import ChartModel
from chart_view import ChartPage, NavView
from cadbiom_gui.gt_gui.graphics.drawing_style import PlainDrawing, NavDrawing
from cadbiom_gui.gt_gui.chart_controler import ChartControler, NavControler
from layout import LayoutVisitor


class EditMVC(object):
    """Editor object composed of a MVC pattern

    Each instance of EditMVC is corresponding to an independant tab
    in the Cadbiom GUI.

    We use the following classes:
        ChartModel: Model
        ChartControler: Model controller
        NavControler: Overview controller

        ChartPage: Graph editor container
        NavView: Graph editor overview
    """

    def __init__(self, chart, model_name, model=None, layout=False):
        """
        :param chart: The charter using the edit mvc
        :param model_name: Name of the model
        :key model: (optional) Chart model
        :key layout: (optional) Flag to activate the default layout (hierarchical_LR).
            Default False.
        :type chart: <Charter>
        :type model_name: <str>
        :type model: <ChartModel>
        :type layout: <boolean>
        """
        self.charter = chart
        self.ident = chart.ei_id
        chart.ei_id = chart.ei_id + 1
        if model:
            self.model = model
        else:
            self.model = ChartModel(model_name)
        self.controler = ChartControler(self.model, chart.clipboard)
        xor, yor, width, height = chart.graph_notebook.get_allocation()
        # adjustment for notebook (-6, -38)
        self.viewpage = ChartPage(
            width - 6, height - 38, PlainDrawing(), self.controler, self.model
        )
        self.view = self.viewpage.draw
        self.title = self.model.name
        # layout?
        if layout:
            # default layout if necessary
            lvi = LayoutVisitor(self.view, "hierarchical_LR")
            model.accept(lvi)
        # overview navigator
        self.navcontroler = NavControler()
        xor, yor, width, height = chart.overview_window.get_allocation()
        self.nav_view = NavView(
            width, height, NavDrawing(), self.navcontroler, self.model
        )
        self.viewpage.attach(self.nav_view)  # linked to view
        self.navcontroler.attach(self.viewpage)

        # registration of auxiliary windows
        self.aux_win = []
        self.has_display = False

    def change_model(self, new_model):
        """Change the model component"""
        self.clean_subwin()
        self.view.model.detach(self.view)
        # new model
        self.model = new_model
        self.controler.model = self.model
        self.navcontroler.model = self.model
        self.view.model = self.model
        self.nav_view.model = self.model
        self.view.model.attach(self.view)
        self.nav_view.model.attach(self.nav_view)
        self.display()

    def display(self):
        """Update and show"""
        self.view.update()
        self.nav_view.update()
        self.viewpage.notify(None)

    def update(self, node, transition):
        """Used in the observer pattern"""
        self.node = node
        self.transition = transition
        self.view.update()
        self.nav_view.update()
        self.viewpage.notify(None)

    def connect(self, chart):
        """Initialize the GUI when the user switches from an EditMVC to another

        We connect the buttons and of the graphical interface to the components
        of the current EditMVC.
        Then the UI is sync with the current model

        :param chart: Mainwindow class
        :type chart: <Charter>
        """
        # connect buttons to emvc controler
        chart.drawing_buttons_connect()
        # self.tab_button.set_sensitive(True)
        # attach charter to new edit_MVC controler
        self.controler.attach("edit_node", chart)
        chart.chart_info.set_controler(self.controler)
        # Force chart_info to be refreshed
        self.controler.notify("current_change")

        # nav window
        chart.overview_window.get_child().add(self.nav_view)

    def disconnect(self, chart, w_destroy):
        """Inverse connect operation

        :param w_destroy: If true the window and its children are destroyed.
            If False, Subordinate windows are not destroyed.
        :type w_destroy: <boolean>
        """
        if w_destroy:
            chart.drawing_button_disconnect()
            chart.drawing_button_enable()
        else:
            chart.drawing_button_disable()
        self.controler.detach("edit_node", chart)
        # self.tab_button.set_sensitive(False)
        if w_destroy:
            self.clean_subwin()
        chart.overview_window.get_child().remove(self.nav_view)

    def on_zoom_plus(self, widget):
        """Zoom in"""
        self.view.zoom_plus()

    def on_zoom_minus(self, widget):
        """Zoom out"""
        self.view.zoom_minus()

    ## sub windows management ##################################################
    def win_register(self, win):
        """Subordinate windows registration"""
        self.aux_win.append(win)

    def win_remove(self, win):
        """Subordinate windows are destroyed"""
        if win in self.aux_win:
            self.aux_win.remove(win)

    def clean_subwin(self):
        """As it says"""
        for win in self.aux_win:
            win.destroy()
        self.aux_win = []



# -*- coding: utf-8 -*-
# Copyright (C) 2010-2017  IRISA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# The original code contained here was initially developed by:
#
#     Geoffroy Andrieux.
#     IRISA/IRSET
#     Symbiose team
#     IRISA  Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
#
#
# Contributor(s): Michel Le Borgne, Nolwenn Le Meur, Pierre Vignet
"""
Created on Nov 2016

..sectionauthor:: Pierre Vignet <pierre.vignet@irisa.fr>
"""
import sys
import os
import pkg_resources
import gtk
from cadbiom_gui.gt_gui.charter import Charter
from cadbiom import commons as cm

LOGGER = cm.logger()


def build_desktop_shortcut(executable_path):
    """Install a .desktop launcher file in ~/.local/share/applications"""

    if sys.platform != "linux2":
        LOGGER.error("This option is only supported on GNU/Linux platforms.")
        exit(1)

    favicon_path = pkg_resources.resource_filename(
        __name__, "gt_gui/images/favicon.ico"
    )

    desktop_content = """[Desktop Entry]
Comment=Cadbiom GUI launcher
Encoding=UTF-8
Exec={}
Icon={}
Categories=Science;
Name=Cadbiom
Type=Application
Terminal=false
""".format(
        executable_path, favicon_path
    )

    desktop_path = "%s/.local/share/applications/" % os.path.expanduser("~")
    with open(desktop_path + "Cadbiom.desktop", "w") as f_d:
        f_d.write(desktop_content)

    LOGGER.info("Desktop file created in <%s>", desktop_path)
    exit(0)


def main():
    """Launch Cadbiom gui"""

    if "--install_desktop_shortcut" in sys.argv:
        build_desktop_shortcut(sys.argv[0])

    charter = Charter(None)
    charter.show()
    gtk.main()



#
## Filename    : warn.py
## Author(s)   : Michel Le Borgne
## Created     : 11/2009
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
A collection of warning widgets
"""

import pygtk

pygtk.require("2.0")
import gtk


def start_warn(message):
    """Warn for a start"""
    mess = gtk.MessageDialog(
        None, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, message
    )
    mess.run()
    mess.destroy()


def ok_warn(message, color=None):
    """Simple warning before action"""
    mess = gtk.MessageDialog(
        None, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, message
    )
    if color:
        color = gtk.gdk.color_parse(color)
        mess.modify_bg(gtk.STATE_NORMAL, color)
    mess.run()
    mess.destroy()


def cancel_warn(message):
    """Error - imply cancellation of the action"""
    message1 = message + " - CANCELLED"
    mess = gtk.MessageDialog(
        None, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, message1
    )
    mess.run()
    mess.destroy()


def system_warn(message):
    """Fatal error - bug or bd problems"""
    message1 = message + " - BD problem or BUG!"
    mess = gtk.MessageDialog(
        None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, message1
    )
    mess.run()
    mess.destroy()


def confirm(parent, message):
    """Ask for action
    :return: Choice of the user (True/False)
    :rtype: <boolean>
    """
    dialog = gtk.Dialog(
        "Warning",
        parent,
        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
        (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT, gtk.STOCK_OK, gtk.RESPONSE_ACCEPT),
    )
    label = gtk.Label(message)
    dialog.vbox.pack_start(label, True, False, 0)
    label.show()
    response = dialog.run()
    dialog.destroy()
    if response == gtk.RESPONSE_ACCEPT:
        return True
    elif response == gtk.RESPONSE_REJECT:
        return False


class DialogEntry(gtk.Dialog):
    """Custom dialog entry"""

    def __init__(self, mess):
        """
        :param mess: Message to be displayed in the dialog box
        :type mess: <str>
        """
        gtk.Dialog.__init__(
            self, mess, None, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT
        )

        self.set_default_size(350, 75)
        self.entry = gtk.Entry()
        self.vbox.pack_start(self.entry, True, True, 0)
        self.entry.show()

        self.okb = gtk.Button("Ok")
        self.action_area.pack_start(self.okb, False, False, 0)
        self.okb.show()

        self.cancel = gtk.Button("Cancel")
        self.action_area.pack_start(self.cancel, False, False, 0)
        self.cancel.show()



##
## Filename    : displayFact.py
## Author(s)   : Michel Le Borgne
## Created     : 11/2009
## Revision    :
## Source      :
##
## Copyright 2009-2010 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
Widgets for lists of nodes used in Cadbiom Gui
"""
import gtk
import gobject


class ToggleList(gtk.Frame):
    """List view with toggle button (Simple/frontier nodes list)"""

    def __init__(self):
        """Create a view

        Attributes::

            :param list_item: Checkable items in the list.
                Internal structure is a tuple of name and check status
                (<tuple <str>, <boolean>>)
            :type list_item: <gtk.ListStore>
        """

        gtk.Frame.__init__(self)
        self.view = gtk.TreeView()
        self.view.show()
        self.add(self.view)
        self.height = 0

        # model
        self.list_item = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_BOOLEAN)
        self.view.set_model(self.list_item)

        # CellRendererText for names
        self.renderer0 = gtk.CellRendererText()
        self.renderer0.set_property("editable", False)

        # CellRendererToggle for selection
        self.renderer1 = gtk.CellRendererToggle()
        self.renderer1.set_property("activatable", True)
        self.renderer1.connect("toggled", self.toggled_col1_callback)
        # connection of cell renderers to respective columns
        self.column0 = gtk.TreeViewColumn(None, self.renderer1)
        self.column1 = gtk.TreeViewColumn(None, self.renderer0, text=0)
        self.column0.add_attribute(self.renderer1, "active", 1)

        self.column0.set_property("sizing", gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.column1.set_property("sizing", gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.view.append_column(self.column0)
        self.view.append_column(self.column1)

    def toggled_col1_callback(self, cellrenderer, path):
        """Callback for column1, called when toggled event is fired

        :param path: position of the cell in the list_item
        """
        self.list_item[path][1] = not self.list_item[path][1]

    def set_list_select(self, l_keywords):
        """
        initialize from a list of name - all unselected
        @param l_keywords: list<string>
        """
        self.height = 40 * len(l_keywords)
        [self.list_item.append([kwo, False]) for kwo in l_keywords]

    def set_list_pairs(self, l_pairs):
        """
        initialize from a list of pairs - all unselected
        @param l_pairs: list< pair<string,bool> >
        """
        self.height = 40 * len(l_pairs)
        [self.list_item.append([pair[0], pair[1]]) for pair in l_pairs]

    def get_selected_items(self):
        """Retrieve the names of the selected items"""
        return [item[0] for item in self.list_item if item[1]]

    def set_selected_items(self, selected_list):
        """Set the check status of the given items in the GUI

        .. note:: Given items are checked, others are unchecked.
        """
        for item in self.list_item:
            if item[0] in selected_list:
                item[1] = True
            else:
                item[1] = False

    def get_pair_values(self):
        """Retrieve the cas and word associated (pairs)"""
        return [(item[0], item[1]) for item in self.list_item]

    def deselect_all(self):
        """Deselect all items"""
        for item in self.list_item:
            item[1] = False

    def select_all(self):
        """Select all items"""
        for item in self.list_item:
            item[1] = True

    def clear(self):
        """Remove all items"""
        self.refresh([])

    def refresh(self, newlist):
        """Rebuild the list"""
        self.list_item.clear()
        self.set_list_select(newlist)


class ChoiceList(object):
    """Display a list with marks (this class is not used)"""

    def __init__(self, l_names, callback, title):
        self.callback = callback

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_resizable(True)
        self.window.connect("destroy", gtk.main_quit)
        height = gtk.gdk.screen_height()
        height = int(height * 0.85)
        self.window.set_size_request(300, height)
        self.window.set_title(title)

        # view
        self.choice_view = gtk.TreeView()

        # model
        self.choice_list = gtk.ListStore(gobject.TYPE_STRING)
        l_names.sort()
        for name in l_names:
            self.choice_list.append([name])
        self.choice_view.set_model(self.choice_list)

        # CellRendererText for choice names
        self.renderer0 = gtk.CellRendererText()
        self.renderer0.set_property("editable", False)

        # connection of cell renderers to respective columns
        self.column0 = gtk.TreeViewColumn("", self.renderer0, text=0)
        self.choice_view.append_column(self.column0)

        swin = gtk.ScrolledWindow()
        swin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        swin.set_shadow_type(gtk.SHADOW_IN)
        swin.add_with_viewport(self.choice_view)
        self.window.add(swin)

        # control
        self.choice_view.connect("row_activated", self.choice_selection_callback)
        self.window.show_all()

    def choice_selection_callback(self, widget, path, view_column):
        """
        callback called when a choice is done
        """
        if widget == self.choice_view:
            treeselection = widget.get_selection()
            (choice, iter) = treeselection.get_selected()
            choice_name = choice.get_value(iter, 0)
            self.callback(choice_name)
            self.window.destroy()



## Filename    : charter_info.py
## Author(s)   : Geoffroy Andrieux
## Created     : 03/2010
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne, Nolwenn Le Meur
##
"""
Widget for displaying information on the elements of the model

CharterInfo: Contains 3 classes inheriting from :class:`Info`:

:class:`ModelInfo`
:class:`NodeInfo`
:class:`TransInfo`

"""
from __future__ import print_function
import gtk
import pkg_resources
import json

from utils.text_page import TextEditConfig, TextPage
from cadbiom import commons as cm


class CharterInfo(gtk.Frame):
    """Windows for displaying information on the elements of the model

    Contains 3 classes whose role is to update in real time the fields in the GUI
    about the currently selected transition, model or node.

    .. seealso::

        :class:`ModelInfo`
        :class:`NodeInfo`
        :class:`TransInfo`

    This class is directly linked to the observers of the :class:`ChartControler`
    thanks to the signal `current_change`.
    """

    def __init__(self, frame):
        self.frame = frame
        self.controler = None
        self.fact_window = None
        self.model_info = ModelInfo(self)
        self.node_info = NodeInfo(self)
        self.trans_info = TransInfo(self)

        self.current_info = self.model_info

    def set_controler(self, new_controler):
        """Associate a controler in a MVC pattern

        .. note:: attach current_change signal to the given controller.

        :param new_controler: New ChartControler
        :type new_controler: <ChartControler>
        """
        if self.controler:
            self.controler.detach("current_change", self)
        self.controler = new_controler
        self.controler.attach("current_change", self)

    def display(self):
        """
        Show all
        """
        chi = self.frame.get_child()
        if chi:
            self.frame.remove(chi)

        self.frame.add(self.current_info.frame)
        self.frame.show()

    def switch_to(self, new_info):
        """
        callback when info change
        """
        if self.current_info is not None:
            self.current_info.warn_change()

        self.current_info = new_info
        self.display()

    def update(self, node, transition):
        """
        Used if registered as an observer
        """
        if node is None and transition is None:
            self.model_info.update(self.controler)
            self.switch_to(self.model_info)

        elif transition is None:
            if node.is_top_node():
                # Update ModelInfo
                self.model_info.update(self.controler)
                self.switch_to(self.model_info)
            else:
                # Update NodeInfo
                self.node_info.update(node)
                self.switch_to(self.node_info)
        else:
            # Update TransInfo
            self.trans_info.update(transition)
            self.switch_to(self.trans_info)

    def enter_callback(self, widget, entry):
        """
        return the entry
        """
        entry_text = entry.get_text()
        print("Entry contents: %s\n" % entry_text)

    def has_transition(self):
        """
        As it says
        """
        return self.current_info.has_transition()


class Info(object):
    """Abstract class

    Implementing a "notes window" for metadata of transitions and nodes

    Used by::

        :class:`ModelInfo`
        :class:`NodeInfo`
        :class:`TransInfo`
    """

    def warn_change(self):
        """Called when ChartInfo is updated (Ex: save notes)"""
        pass

    def has_transition(self):
        """Return the current transition"""
        pass

    def update(self, controler):
        """Used when registered as an observer"""
        pass

    def display_note_window(self, title, note_text):
        """Display a window with the given text content

        Used to display notes from the selected transitions or nodes.

        .. seealso: :class:`NodeInfo`, :class:`TransInfo`.

        :param title: Title of the window
        :param note_text: Text content to display
        :type title: <str>
        :type note_text: <str>
        """
        self.note_window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        # TEST: Do not save the notes
        # self.note_window.connect("destroy", self.save_note)
        self.note_window.set_title("notes: " + title)
        self.note_window.set_position(gtk.WIN_POS_CENTER)
        self.note_window.set_keep_above(True)
        self.note_window.set_default_size(600, 300)
        self.note_window.connect("key_press_event", self.on_escape)
        # Favicon
        favicon = pkg_resources.resource_filename(
            __name__,
            "images/favicon.ico"
        )
        self.note_window.set_icon_from_file(favicon)

        ed_conf = TextEditConfig()
        self.note_text_page = TextPage(None, ed_conf)
        # TEST: Not editable
        self.note_text_page.write.set_editable(False)
        # Handle JSON data
        try:
            parsed = json.loads(note_text)
            note = json.dumps(parsed, indent=2, sort_keys=True)
        except ValueError:
            note = note_text

        self.note_text_page.set_text(note)
        self.note_window.add(self.note_text_page)
        self.note_window.show_all()

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.note_window.destroy()

    def save_note(self, widget):
        """Save the current text modified in display_note_window"""
        raise NotImplementedError("Abstract method")


class ModelInfo(Info):
    """Object used as a model - store information for CharterInfo"""

    def __init__(self, cin):
        template = pkg_resources.resource_filename(
            __name__, "chart_glade/model_info.glade"
        )
        self.page = gtk.glade.XML(template)
        wid = self.page.get_widget("model_frame")
        self.frame = wid.get_child()
        wid.remove(self.frame)
        self.info = cin

        self.name = self.page.get_widget("model_entry")
        self.name.connect("changed", self.set_model)
        self.name.set_editable(True)

        self.nb_nodes = self.page.get_widget("node_entry")
        self.nb_nodes.set_editable(False)
        self.nb_nodes.set_sensitive(False)

        self.nb_trans = self.page.get_widget("trans_entry")
        self.nb_trans.set_editable(False)
        self.nb_trans.set_sensitive(False)

    def set_model(self, widget):
        """Called when the name of the model has been modified by a user
        - set the name of the model,
        - then notify the model.
        """
        self.model.name = self.name.get_text()
        self.model.notify()

    def update(self, controler):
        """
        when used as an observer
        """
        self.model = controler.model
        self.name.set_text(controler.model.name)
        nodes = str(len(self.model.simple_node_dict.keys()))
        trans = str(len(self.model.transition_list))
        self.nb_nodes.set_text(nodes)
        self.nb_trans.set_text(trans)


class NodeInfo(Info):
    """A component of CharterInfo used for nodes

    .. TODO:: Display node notes; move notes code from TransInfo to Info
    """

    def __init__(self, cin):
        template = pkg_resources.resource_filename(
            __name__, "chart_glade/node_info.glade"
        )
        self.page = gtk.glade.XML(template)
        wid = self.page.get_widget("node_frame")
        self.frame = wid.get_child()
        wid.remove(self.frame)
        self.info = cin

        self.name = self.page.get_widget("node_entry")
        self.name.connect("changed", self.set_node)

        # Button
        self.show_notes_button = self.page.get_widget("button_metadata")
        self.show_notes_button.connect("clicked", self.set_note)

    def set_node(self, widget):
        """Called when the node has been modified by a user
        - set the node settings,
        - then notify the model.
        """
        self.node.set_name(self.name.get_text())
        self.node.model.notify()

    def update(self, node):
        """
        Used when registered as an observer
        """
        if node:
            self.node = node
            self.name.set_text(node.name)

    def set_note(self, widget):
        """Associate a text note to a node

        Use the method of the parent class :class:`Info` to display the note.
        """
        self.display_note_window(self.node.name, self.node.note)


class TransInfo(Info):
    """A component of CharterInfo used for transitions"""

    def __init__(self, cin):

        template = pkg_resources.resource_filename(
            __name__, "chart_glade/trans_info.glade"
        )
        self.page = gtk.glade.XML(template)
        wid = self.page.get_widget("trans_frame")
        self.frame = wid.get_child()
        wid.remove(self.frame)
        self.info = cin  # owner
        self.note_window = None

        # Text entries
        self.name = self.page.get_widget("trans_entry")
        self.name.set_editable(False)
        self.evt = self.page.get_widget("evt_entry")
        self.cond = self.page.get_widget("cond_entry")

        # Buttons
        self.show_notes_button = self.page.get_widget("button_metadata")
        # Attached to Charter.on_display_states via clicked event
        self.show_influencing_nodes_button = self.page.get_widget(
            "button_influencing_nodes"
        )

        # Tooltips
        self.tooltips = gtk.Tooltips()
        self.tooltips.set_tip(self.show_notes_button, "Show the transition metadata")
        self.tooltips.set_tip(
            self.show_influencing_nodes_button,
            "Show the nodes influencing this transition",
        )
        self.tooltips.enable()
        self.tooltips.set_delay(cm.TOOLTIPS_DELAY)

        # User can change all the entries
        self.name.connect("changed", self.set_trans)
        self.evt.connect("changed", self.set_trans)
        self.cond.connect("changed", self.set_trans)
        self.show_notes_button.connect("clicked", self.set_note)

    def set_trans(self, widget):
        """Called when the transition has been modified by a user
        (and after TransInfo update)

        - set the transition settings,
        - then notify the model.
        """
        if self.lock:
            return

        self.trans.set_name(self.name.get_text())
        self.trans.set_event(self.evt.get_text())
        self.trans.set_condition(self.cond.get_text())
        self.trans.ori.model.notify()

    def warn_change(self):
        """
        As it says
        """
        # TEST: Do not save the notes
        # self.save_note(None)
        pass

    def update(self, trans):
        """
        When registered as an observer
        """
        if trans:
            self.lock = True
            self.trans = trans  # for fact association management
            self.name.set_text(trans.ori.name + "->" + trans.ext.name)
            self.evt.set_text(trans.event)
            self.cond.set_text(trans.condition)
            self.lock = False

    def has_transition(self):
        """
        As it says
        """
        return self.trans

    def set_note(self, widget):
        """Associate a text note to a transition

        Use the method of the parent class :class:`Info` to display the note.
        """
        self.display_note_window(
            self.trans.ori.name + "->" + self.trans.ext.name, self.trans.note
        )

    def save_note(self, widget):
        """
        register the note (deactivated for now)
        """
        if self.note_window:
            text = self.note_text_page.get_text()
            self.trans.note = text
            self.trans.ori.model.modified = True
            self.note_window.destroy()
            self.note_window = None



# -*- coding: utf-8 -*-
## Filename    : chart_controler.py
## Author(s)   : Michel Le Borgne
## Created     : 4/3/2010
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
##     http:
##     mailto:
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""Main GUI controllers + auxiliary class (ChartClipboard)

- :class:`ChartClipboard`: A clipboard to handle the copy of nodes through models
- :class:`ChartControler`: A controler for graphical views
- :class:`NavControler`: A controler for navigation view (overview section in the GUI)

"""
# Standard imports
from __future__ import print_function
import itertools as it
from string import ascii_uppercase
from math import sqrt
import re

# Custom imports
import gtk
from gtk.gdk import (
    Cursor,
    ARROW,
    BOTTOM_LEFT_CORNER,
    BOTTOM_RIGHT_CORNER,
    TOP_LEFT_CORNER,
    TOP_RIGHT_CORNER,
    LINE_ON_OFF_DASH,
)

# Cadbiom imports
from cadbiom_gui.gt_gui.graphics.drawing_style import Arrow
from cadbiom.models.guard_transitions.chart_model import CMacroNode, CNode, CSimpleNode
from cadbiom import commons as cm

LOGGER = cm.logger()


class ChartClipboard(object):
    """A clipboard used by ChartControler to handle the copy of nodes through models.
    """

    def __init__(self):
        self.clip_node = None

    def put_node(self, node):
        """register a chart model node in the clipboard"""
        self.clip_node = node

    def get_node(self):
        """retrieve the current node in the clipboard"""
        node = self.clip_node
        self.clip_node = None
        return node

    def has_element(self):
        """test if the clipboard has a registered node"""
        return self.clip_node is not None

    def has_macro(self):
        """test if the clipboard has a registered macro node"""
        return self.clip_node.is_macro() if self.clip_node else False


class ChartControler(object):
    """Implement a controler for graphical views

    Used by::

        - the main graph editor widget as this
        - NavControler for the overview widget through inheritance

    Signals::

        - current_change: Inform CharterInfo that the currently selected transition,
        model or node has changed. Used to dispatch related information in the GUI.
        - edit_node: Inform Charter that a MacroNode will be edited.

    TODO: just show metadata on double click on another item
    => add a signal to inform CharterInfo objects

    :param model: Current chart model
    :param mouse_role: Mouse action "select", "resizing", "moving"
    :param current_node: Currently selected node
    :param current_node_center: Tuple of coordinates (x, y) of the current node
    :param current_handle: Id of the node corner currently selected for resizing
        (only for MacroNodes)
    :param current_transition: Currently selected transition
    :param m_vscreen_coord: Mouse virtual screen coordinates (xloc, yloc)
    :param lastx: coordinates of last click in view
    :param lasty: coordinates of last click in view
    :param clipboard: Clipboard object used to copy nodes through models
    :param drawing_style:
    :param signal_dict: Structure to assign observers to each signal of ChartControler
    :param gen_name: Generator of unique names for new nodes
    :param node_copy_count: Generator of ints used in the naming of copied nodes

    :type model: <ChartModel>
    :type mouse_role: <str>
    :type current_node: <CNode>
    :type current_node_center: <tuple <float>,<float>>
    :type current_handle: <int>
    :type current_transition: <CTransition>
    :type m_vscreen_coord: <tuple <float>,<float>>
    :type lastx: <float>
    :type lasty: <float>
    :type clipboard: <ChartClipboard>
    :type drawing_style:
    :type signal_dict: <dict <str>:<list>>
    :type gen_name: <generator>
    :type node_copy_count: <generator>
    """

    cursors = []
    cursors.append(Cursor(ARROW))
    cursors.append(Cursor(TOP_LEFT_CORNER))
    cursors.append(Cursor(TOP_RIGHT_CORNER))
    cursors.append(Cursor(BOTTOM_RIGHT_CORNER))
    cursors.append(Cursor(BOTTOM_LEFT_CORNER))
    LSIGNALS = ["current_change", "edit_node"]

    def __init__(self, model, clipboard):
        self.model = model
        # mouse role
        self.mouse_role = "select"
        # Set default node
        self.current_node = self.model.get_root()
        # Note: current_node_center should be initialized to None
        # only if current_node is also None...
        self.current_node_center = None
        self.current_handle = 0
        self.current_transition = None
        # coordinate of last click in virtual screen 1x1
        self.m_vscreen_coord = None
        # coordinates of last click in view
        self.lastx = 0
        self.lasty = 0
        self.clipboard = clipboard
        self.drawing_style = None
        self.signal_dict = {signal: list() for signal in self.LSIGNALS}

        # Generator for the auto-naming of new nodes
        self.gen_name = self.nodes_names_generator()
        # Generator for the auto-naming of copied nodes
        self.node_copy_count = it.count(1)

    def attach(self, signal, obs):
        """Register an observer for the given signal

        :param signal: Name of the signal ("current_change" or "edit_node")
        :param obs: The observer
        :type signal: <str>
        :type obs: <CharterInfo> or <Charter>
        """
        lobs = self.signal_dict[signal]
        if not obs in lobs:
            lobs.append(obs)

    def detach(self, signal, obs):
        """Remove an observer for the given signal

        :param signal: Name of the signal ("current_change" or "edit_node")
        :param obs: The observer
        :type signal: <str>
        :type obs: <CharterInfo> or <Charter>
        """
        self.signal_dict[signal].remove(obs)

    def notify(self, signal):
        """Emit a signal subsequently to a mouse action

        - current_change: Tell CharterInfo that the current selection has changed
        - edit_node: Tell Charter that the current MacroNode will be edited

        :param signal: name of the signal
        :type signal: <str>
        """
        LOGGER.debug("ChartControler notify signal: %s to %s", signal, self.signal_dict[signal])

        lobs = self.signal_dict[signal]
        if signal == "current_change":
            # Tell CharterInfo that the selection has changed
            for obs in lobs:
                obs.update(self.current_node, self.current_transition)
        elif signal == "edit_node":
            # Tell Charter that the MacroNode will be edited
            for obs in lobs:
                obs.update(self.current_node)
        else:
            raise Exception("ChartControler: Unknown signal")

    def set_view(self, view):
        """
        As it says
        """
        self.view = view

    def set_mouse_role(self, role):
        """
        As it says
        """
        self.mouse_role = role

    def on_button_press(self, widget, event):
        """Callback when a mouse button is pressed on the DrawingArea
        (contained in the ChartView object)

        - Detect the current element under the mouse (node or transition)
        - Select it, and notify observers
        - Interpret mouse actions according to the current mouse role::

            - Double click: Edit the node if it is a MacroNode
            - Left click: Resize (a MacroNode) or move
            - Right click: Open context menu

        TODO: just show metadata on double click on another item

        :param widget: gtk DrawingArea that emit the event
        :param event: gdk mouse event with attributes (x, y)
        :type widget: <NavView> or <ChartView>
        :type event: <gtk.gdk.Event>
        """
        self.lastx = event.x
        self.lasty = event.y

        # cursor coordinates in virtual screen 1x1
        xloc = event.x / self.view.draw_width
        yloc = event.y / self.view.draw_height
        self.m_vscreen_coord = (xloc, yloc)

        # Get current object under the mouse
        (node, handle, center, trans) = self.model.find_element(
            self.m_vscreen_coord, self.drawing_style
        )

        # Note: We are always in a node, which one?
        if node == self.model.get_root():
            # we found the root of the sub_model - no resizing allowed
            handle = 0
        self.current_handle = handle


        if self.current_node != node:
            # Node detected, update current node and notify observers
            if self.current_node:
                self.current_node.selected = False

            self.current_node = node
            node.selected = True
            self.current_node_center = center

            # Notify that we have found something (red color for selected element)
            self.notify("current_change")
            # For deselection, see at the end of the transitions block
            self.model.notify()

        if self.current_transition != trans:
            # Transition detected, update current node/transition and notify observers
            if self.current_transition:
                self.current_transition.selected = False

            if trans:
                node.selected = False
                trans.selected = True
                self.current_transition = trans
            else:
                self.current_transition = None

            # Notify that we have found something (red color for selected element)
            self.notify("current_change")
            # TODO: Don't know why, but the transitions requires
            # a notification to the model (this is not the case for nodes that
            # can be selected without notify but still require notify to deselect..)
            self.model.notify()


        # Action (depending on mouse button and event type)
        if event.type == gtk.gdk._2BUTTON_PRESS:
            # Double click
            self.mouse_role = "select"
            if self.current_node and self.current_node.is_macro():
                self.notify("edit_node")

            # TODO: just show metadata on double click on another item

        if event.button == 1:
            # Left click
            if self.mouse_role == "select" and node:
                # got a node
                if handle != 0:
                    # we get a handle => resizing
                    self.mouse_role = "resizing"
                else:
                    self.mouse_role = "moving"
        elif event.button == 3:
            # Right click
            self.context_menu(event)

    def context_menu(self, event):
        """Menu appearing on right click"""
        menu = gtk.Menu()
        node = self.current_node
        transition = self.current_transition
        # Create menu-item copy for nodes only
        if not transition:
            class_name = node.__class__.__name__
            action = "Copy the current {}".format(
                "model" if class_name == "CTopNode" else class_name[1:]
            )

            menu_items = gtk.MenuItem(action)
            menu.append(menu_items)
            menu_items.connect("activate", self.menuitem_response, action, node)
            menu_items.show()
        # Create menu-item remove (do not remove top node)
        if transition or node != self.model.get_root():
            action = "Remove"
            menu_items = gtk.MenuItem(action)
            menu.append(menu_items)
            menu_items.connect(
                "activate", self.menuitem_response, action, transition or node
            )
            menu_items.show()
        # Create menu-item paste
        action = "Paste"
        menu_items = gtk.MenuItem(action)
        menu.append(menu_items)
        if self.clipboard.has_element():
            menu_items.connect("activate", self.menuitem_response, action, node)
        else:
            menu_items.set_sensitive(False)
        menu_items.show()
        # Create menu item paste component
        action = "Paste the model components"
        menu_items = gtk.MenuItem(action)
        menu.append(menu_items)
        if self.clipboard.has_macro():
            menu_items.connect("activate", self.menuitem_response, action, node)
        else:
            menu_items.set_sensitive(False)
        menu_items.show()

        menu.popup(None, None, None, 3, event.time)

    def menuitem_response(self, widget, action, item):
        """Callback for the context menu described in context_menu()

        :param action: Option selected
        :param item: Current selected node
        :param widget: Widget that emitted the event
        :type action: <str>
        :type item: <CTransition> or <CNode>
        """

        def update_nodes_model(sub_nodes):
            """
            - Set recursively the model for all the given nodes
            - Rename the nodes and make them as unique as possible
            - Update the transitions of the destination model
            - Update the nodes of the destination model

            .. TODO: Do not rename nodes/transitions if they are copied on a new model.
                Follow copy_number variable...
            """
            for sub_node in sub_nodes:
                if isinstance(sub_node, CMacroNode):
                    # MacroNode => recursive process
                    # LOGGER.debug("menuitem_response:: MacroNode here: %s", sub_node)
                    update_nodes_model(sub_node.sub_nodes)

                    # Update the transitions of the destination MODEL
                    # not the parent node himself, this will eventually be done later
                    node.model.transition_list += list(it.chain(*sub_node.transitions))

                assert sub_node.model is not None

                # LOGGER.debug("menuitem_response:: update model for %s", sub_node)
                sub_node.model = node.model

                ## New model ? do not rename...
                # Naming
                if sub_node.is_start() or sub_node.is_trap():
                    # Renumber start and trap nodes (add a number)
                    sub_node.name += str(node.start_trap_nodes_count)
                    node.start_trap_nodes_count += 1
                else:
                    # Rename all other nodes (names must be as unique as possible)
                    sub_node.name += "_" + str(copy_number)

                # Update model (do not ask me why this duplication of dict...)
                # simple_node_dict is for the GUI: list only CSimpleNodes
                # Start/Trap/Input/Perm/Macro/Top are only in node_dict
                if isinstance(sub_node, CSimpleNode) and not isinstance(
                    sub_node, CMacroNode
                ):
                    node.model.simple_node_dict[sub_node.name] = sub_node
                node.model.node_dict[sub_node.name] = sub_node

        def update_transitions(child_node):
            """
            Update the macro_node of the new transitions
            Update the transitions of the destination MODEL with the child_node ones
            Rename transitions
            Update names of nodes in conditions

            .. note:: This will also update child_node transitions (side effect)
            .. note:: Must be before new_transitions.update
                in order to avoid double counting of transitions
                (node.transitions already contains current model transitions
                and we don't want to modify them)
            """
            for transitions in child_node.transitions:
                # Model
                node.model.transition_list += transitions
                # MacroNode
                for transition in transitions:
                    transition.macro_node = node
                    # print("transition", transition)
                    # print("macro node", transition.macro_node)
                    # print("ori out:", transition.ori.outgoing_trans)
                    # print("ori in:", transition.ori.incoming_trans)
                    # print("event:", transition.event)
                    # print("influencing places:", transition.get_influencing_places())

                    # Rename nodes in the condition
                    # Ex: "A and not D) or (A or B) and (D and not A"
                    # Becomes: 'A and not D) or (A_1 or B) and (D and not A_1'
                    for place in transition.get_influencing_places():
                        transition.condition = re.sub(
                            r"([( ]|^)(%s)([) ]|$)" % place,
                            r"\g<1>%s\g<3>" % (place + "_" + str(copy_number)),
                            transition.condition
                        )

                    # Rename the transition
                    transition.event += "_" + str(copy_number)

            # Clean transitions duplicates added by update_nodes_model()
            node.model.transition_list = list(set(node.model.transition_list))


        if "Copy" in action:
            if not isinstance(item, CNode):
                return
            # Only for nodes
            child_node = item.copy()
            self.clipboard.put_node(child_node)

        elif action == "Remove":
            self.remove_node_or_transition(item)

        elif action == "Paste":
            if self.clipboard.has_element() and isinstance(item, CMacroNode):
                # Add a CMacroNode/CTopNode to the current model
                node = item
                child_node = self.clipboard.get_node()
                # Prepare naming of copied nodes
                copy_number = next(self.node_copy_count)

                # Update the parent and the model of child_node
                node.sub_nodes.append(child_node)
                child_node.father = node
                child_node.set_model(node.model)
                child_node.name += "_" + str(copy_number)

                # print("current child node model updated:", child_node.model)
                # print("current child node type:", type(child_node))

                if isinstance(child_node, CMacroNode):
                    # child_node is a complex node, we have to recursively
                    # update its content

                    # Update the model of the children of child_node (MacroNode)
                    update_nodes_model(child_node.sub_nodes)

                    # Update the macro_node of the new transitions
                    # Update the transitions of the destination MODEL
                    update_transitions(child_node)

                    # Add transitions to the current top/macro node ?
                    # => NO!! Transitions must remain in their complex node
                    # of belonging; nevertheless they are already referenced
                    # in the model thanks to update_transitions()

                # Coordinates in new node
                child_node.xloc = self.m_vscreen_coord[0]
                child_node.yloc = self.m_vscreen_coord[1]

                # Refresh ui
                node.model.modified = True
                node.model.notify()

        elif action == "Paste the model components":
            if self.clipboard.has_element() and isinstance(item, CMacroNode):
                # Add the content of a CMacroNode/CTopNode to the current model
                child_node = self.clipboard.get_node()

                if not isinstance(child_node, CMacroNode):
                    return

                node = item
                # Prepare naming of copied nodes
                copy_number = next(self.node_copy_count)

                # Add nodes to the current top/macro node
                # For direct chilren: Update their parent
                for snode in child_node.sub_nodes:
                    node.sub_nodes.append(snode)
                    snode.father = node

                # Update the model attr of the children of child_node
                # Note: Don't update the model of child_node, since it is
                # not used (we use only its children)
                update_nodes_model(child_node.sub_nodes)

                # Update the macro_node of the new transitions
                # Update the transitions of the destination MODEL
                update_transitions(child_node)

                # Add transitions to the current top/macro node
                # Note: Don't overwrite data of the current MacroNode => update
                # print("Transitions before:", node.new_transitions)
                node.new_transitions.update(child_node.new_transitions)
                # print("Transitions copied:", child_node.new_transitions)

                # Refresh ui
                node.model.modified = True
                node.model.notify()

    def remove_node_or_transition(self, item):
        """Remove the given transition or the given node

        Called by Charter object when Delete key is pressed, or when the Remove
        option is selected on the the context-menu hover an item on the DrawingArea.

        :param item:
        :type item: <CTransition> or <CNode> but not <CTopNode>
        """
        # Remove the item
        item.remove()

        if not isinstance(item, CNode):
            # Transition
            self.current_transition = None
            # TODO: Don't know why, but the deletion of transitions requires
            # a notification to the model (this is not the case for nodes)
            self.model.notify()

        # Reset selection
        self.current_node = None
        self.current_node_center = None
        # Current handle is now the TopNode
        self.current_handle = 0

        # Refresh ChartInfo in order to display actualized data about the model
        self.notify("current_change")

    def on_button_release(self, widget, event):
        """Callback when a mouse button is released on the DrawingArea
        (in the ChartView object)

        :param widget: gtk DrawingArea that emit the event
        :param event: gdk mouse event with attributes (x, y)
        :type widget: <NavView> or <ChartView>
        :type event: <gtk.gdk.Event>
        """
        # rubout last click characteristics
        if self.mouse_role == "select":
            return

        elif self.mouse_role == "resizing" or self.mouse_role == "moving":
            self.view.window.set_cursor(ChartControler.cursors[0])
            self.mouse_role = "select"

        elif self.mouse_role == "new_trans" and self.current_node:
            self.new_transition(event.x, event.y)

        elif self.current_node and self.current_node.is_macro():
            # assume node creation
            self.new_node(self.mouse_role)

    def new_node(self, node_type):
        """Creation of a new node

        Called during the :meth:`on_button_release` callback.

        @param node_type: type of the node (string)
        """
        xnode = self.m_vscreen_coord[0]
        ynode = self.m_vscreen_coord[1]
        if node_type == "simple":
            self.current_node = self.current_node.add_simple_node(
                next(self.gen_name), xnode, ynode
            )
        elif node_type == "macro":
            self.current_node = self.current_node.add_macro_subnode(
                next(self.gen_name), xnode, ynode, 0.25, 0.25
            )
        elif node_type == "start":
            self.current_node = self.current_node.add_start_node(xnode, ynode)
        elif node_type == "trap":
            self.current_node = self.current_node.add_trap_node(xnode, ynode)
        elif node_type == "perm":
            self.current_node = self.current_node.add_perm_node(
                next(self.gen_name), xnode, ynode
            )
        elif node_type == "input":
            self.current_node = self.current_node.add_input_node(
                next(self.gen_name), xnode, ynode
            )
        elif node_type == "env":
            self.current_node = self.current_node.add_env_node(
                next(self.gen_name), xnode, ynode, 0.25, 0.25
            )
        else:  # bug!
            raise TypeError("new_node: UNKNOWN TYPE: %s" % node_type)

        self.current_transition = None
        self.current_handle = 0
        self.m_vscreen_coord = (0.0, 0.0)
        self.notify("current_change")
        self.view.window.set_cursor(ChartControler.cursors[0])
        self.mouse_role = "select"

    def nodes_names_generator(self):
        """Return a generator of names for new nodes.

        Names are generated in lexicographic order and we try to avoid names
        that are already in the model (this is important in order to avoid
        overwriting of nodes).
        """
        for size in it.count(start=1):
            for tpl in it.combinations(ascii_uppercase, size):
                name = "".join(tpl)
                if name in self.model.node_dict:
                    continue
                yield name

    def new_transition(self, xmo, ymo):
        """Create a new transition

        Called during the :meth:`on_button_release` callback.

        @param xmo, ymo: int mouse screen coordinates
        """
        # are we in a node?
        xloc = xmo / self.view.draw_width
        yloc = ymo / self.view.draw_height
        #        w_coef = 1.0/self.view.draw_width
        #        h_coef = 1.0/self.view.draw_height
        (node, handle, center, trans) = self.model.find_element(
            (xloc, yloc), self.drawing_style
        )
        if not node:  # no extremity (never happens)
            self.view.window.set_cursor(ChartControler.cursors[0])
            self.mouse_role = "select"
            self.model.notify()  # rubout the draft transition
            return
        # are we in same container
        if self.current_node.father != node.father:
            self.view.window.set_cursor(ChartControler.cursors[0])
            self.mouse_role = "select"
            self.model.notify()  # rubout the draft transition
            return
        # we are in extremity node - current node is origin node
        # is origin node correct and extremity node correct
        if self.current_node.is_for_origin() and node.is_for_extremity():
            trans = self.current_node.father.add_transition(self.current_node, node)
            if trans:
                self.current_node.selected = False
                self.current_node = None
                self.m_vscreen_coord = (0.0, 0.0)
                self.current_transition = trans
                trans.selected = True
                self.notify("current_change")
                self.model.notify()
                self.view.window.set_cursor(ChartControler.cursors[0])
                self.mouse_role = "select"
            else:
                self.view.window.set_cursor(ChartControler.cursors[0])
                self.mouse_role = "select"
                self.model.notify()  # rubout the draft transition
        else:
            self.view.window.set_cursor(ChartControler.cursors[0])
            self.mouse_role = "select"
            if self.current_node:
                self.current_node.selected = False
            self.model.notify()  # rubout the draft transition
            return

    def on_motion_notify(self, widget, event):
        """Callback moving the cursor in the DrawingArea
        (in the ChartView object)

        :param widget: gtk DrawingArea that emit the event
        :param event: gdk mouse event with attributes (x, y)
        :type widget: <NavView> or <ChartView>
        :type event: <gtk.gdk.Event>
        """
        swi = self.view.draw_width
        she = self.view.draw_height

        # transform mouse coordinates to virtual screen 1.0 x 1.0
        xvirt = event.x / swi
        yvirt = event.y / she

        if self.mouse_role == "select":
            (node, handle, center, trans) = self.model.find_element(
                (xvirt, yvirt), self.drawing_style
            )
            if node == self.model.get_root():
                handle = 0
            # detect a new handle
            if handle != self.current_handle:
                self.current_handle = handle
                self.view.window.set_cursor(ChartControler.cursors[handle])

        elif self.mouse_role == "resizing":
            self.current_node.resize(
                xvirt, yvirt, self.current_handle, swi, she, self.model.get_root()
            )

        elif self.mouse_role == "moving":
            v_dx = xvirt - self.m_vscreen_coord[0]
            v_dy = yvirt - self.m_vscreen_coord[1]
            v_size = self.current_node.accept(self.drawing_style)
            self.current_node.move(v_dx, v_dy, v_size, self.model.get_root())
            self.m_vscreen_coord = (xvirt, yvirt)
        elif self.mouse_role == "new_trans":
            if self.current_node:
                self.draft_transition(event.x, event.y)

    def draft_transition(self, xmo, ymo):
        """
        draw a transition arrow in dotted line
        @param xmo,ymo: mouse coordinates
        """
        view = self.view
        self.model.notify()
        xr1 = int(self.current_node_center[0] * self.view.draw_width)
        yr1 = int(self.current_node_center[1] * self.view.draw_height)

        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_line_attributes(1, LINE_ON_OFF_DASH, 0, 0)
        # line
        pixmap.draw_line(grc, xr1, yr1, int(xmo), int(ymo))
        # arrow
        unx = xmo - xr1
        uny = ymo - yr1
        norm = sqrt(unx ** 2 + uny ** 2)
        if norm != 0:
            unx = unx / norm
            uny = uny / norm
            arr = Arrow()
            arr.draw(view, (unx, uny), (xmo, ymo))
        grc = view.window.new_gc()
        view.window.draw_drawable(
            grc, view.pixmap, 0, 0, 0, 0, view.draw_width, view.draw_height
        )

    ## API for communication with charter ######################################
    def set_action_select(self, widget):
        """
        As it says
        """
        self.mouse_role = "select"

    def set_action_new_simple_node(self, widget):
        """
        As it says
        """
        self.mouse_role = "simple"

    def set_action_new_macro_node(self, widget):
        """
        As it says
        """
        self.mouse_role = "macro"

    def set_action_new_start_node(self, widget):
        """
        As it says
        """
        self.mouse_role = "start"

    def set_action_new_trap_node(self, widget):
        """
        As it says
        """
        self.mouse_role = "trap"

    def set_action_new_input_node(self, widget):
        """
        As it says
        """
        self.mouse_role = "input"

    def set_action_new_transition(self, widget):
        """
        As it says
        """
        if self.current_node:
            self.current_node.selected = False
        self.current_node = None
        self.current_handle = 0
        self.m_vscreen_coord = None
        if self.current_transition:
            self.current_transition.selected = False
        self.current_transition = None
        self.mouse_role = "new_trans"

    def set_action_new_perm_node(self, widget):
        """
        As it says
        """
        self.mouse_role = "perm"


class NavControler(object):
    """Controler for navigation view (overview section in the GUI)
    Implement a controler for navigation views
    """

    def __init__(self):
        self.lastx = 0  # coordinates of last click in view
        self.lasty = 0
        self.in_ret = False
        self.obs = []

    def set_view(self, view):
        """
        As it says
        """
        # attribute view is assigned when a view is created
        self.view = view

    def attach(self, obs):
        """
        observer management
        """
        if not obs in self.obs:
            self.obs.append(obs)

    def detach(self, obs):
        """
        observer management
        """
        self.obs.remove(obs)

    def notify(self, depx, depy):
        """
        observer management
        """
        for obs in self.obs:
            obs.update(depx, depy)

    def on_button_press(self, widget, event):
        """
        action when a mouse button is pressed
        """
        self.lastx = event.x
        self.lasty = event.y
        cond = self.view.x_ret <= self.lastx
        cond = cond and (self.lastx <= (self.view.x_ret + self.view.w_ret))
        if cond:
            cond2 = self.view.y_ret <= self.lasty
            vvv = self.view.y_ret + self.view.h_ret
            cond2 = cond2 and (self.lasty <= vvv)
            if cond2:
                self.in_ret = True

    def on_button_release(self, widget, event):
        """
        action when a mouse button is released
        """
        self.in_ret = False

    def on_motion_notify(self, widget, event):
        """
        callback
        """
        if self.in_ret:
            depx = event.x - self.lastx
            depy = event.y - self.lasty
            alloc = widget.get_allocation()
            # limits
            vv1 = self.view.x_ret + depx + self.view.w_ret
            vv2 = self.view.x_ret + depx
            if (vv1 > alloc.width) or (vv2 < 0):
                depx = 0
            vv1 = self.view.y_ret + depy + self.view.h_ret
            vv2 = self.view.y_ret + depy
            if (vv1 > alloc.height) or (vv2 < 0):
                depy = 0
            self.lastx = self.lastx + depx
            self.lasty = self.lasty + depy
            # reticule move
            if (not depx == 0) or (not depy == 0):
                depx = float(depx) / alloc.width
                depy = float(depy) / alloc.height
                self.notify(depx, depy)



# -*- coding: utf-8 -*-
## Filename    : chart_checker_controler.py
## Author(s)   : Michel Le Borgne
## Created     : 05/2011
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
Collection of widgets for properties checking

Classes:

- :class:`ChartChecker`: GUI interface for query checking
- :class:`QueryCheckingForm`: Main form of ChartChecker Window
- :class:`SolutionWindow`: Show solutions found by the solver
- :class:`PropertyVisitor`: Used to parse places in a start property and force
  their use by the simulator
- :class:`InhibWindow`: Show inhibitors - Not used anymore

"""
from collections import defaultdict
import traceback
import pkg_resources

import gtk

from cadbiom.models.clause_constraints.mcl.MCLAnalyser import MCLAnalyser
from cadbiom_gui.gt_gui.chart_simulator.chart_simul_controler import ChartSimulControler
from cadbiom_gui.gt_gui.chart_simulator.chart_chrono import ChartChrono
from cadbiom_gui.gt_gui.utils.warn import ok_warn, cancel_warn, confirm
from cadbiom_gui.gt_gui.utils.reporter import CompilReporter
from cadbiom.models.guard_transitions.simulator.chart_simul import ChartSimulator
from cadbiom.models.clause_constraints.mcl.MCLQuery import MCLSimpleQuery

from cadbiom.models.biosignal.sig_expr import SigBinExpr
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond

from cadbiom import commons as cm

LOGGER = cm.logger()


class ChartChecker(object):
    """Provides a GUI interface for query checking"""

    def __init__(self, emvc, reporter, parent=None):
        self.__emvc = emvc  # edit mvc - link with charter
        self.__reporter = reporter
        self.__occ_form = QueryCheckingForm(emvc, reporter)
        self.__current_form = self.__occ_form

        # graphical interface
        self.__main_window = gtk.Window()
        self.__main_window.set_title("Property check: " + self.__emvc.model.name)
        self.__main_window.connect("destroy", self.__on_destroy)
        self.__main_window.set_position(gtk.WIN_POS_CENTER)
        self.__main_window.resize(600, 300)

        if parent:
            # Set window above all windows
            self.__main_window.set_transient_for(parent.main_window)

        # Event on_escape key pressed
        self.__main_window.connect("key_press_event", self.on_escape)

        # register as auxiliary window
        self.__emvc.win_register(self)

        # display
        self.__main_window.add(self.__current_form.get_frame())
        self.__main_window.show_all()

    def __on_destroy(self, widget):
        """
        delete window and dependant sub widgets
        """
        self.__current_form.clean_subwin()
        if self.__main_window:
            self.__main_window.destroy()

    def destroy(self):
        """
        call back adaptor
        """
        self.__on_destroy(None)

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.__main_window.destroy()


class QueryCheckingForm(object):
    """General simple query checking; used by ChartChecker

    .. TODO::
        Exemple example_model.bcx:
        Final property: "C"
        Start property: A and G

        Pourquoi G se retrouve dans les places frontières des solutions et pas A ?
        Car c'est déjà une frontière et pas A...
        => bug ?
    """

    def __init__(self, emvc, er_rep):
        self.__emvc = emvc
        self.__model = emvc.model  # chart model
        self.__mcla = None  # clause constraint model analyser
        self.__error_reporter = er_rep
        template = pkg_resources.resource_filename(
            __name__, "../chart_glade/occurence_form.glade"
        )
        self.__wtree = gtk.glade.XML(template)
        self.__window = self.__wtree.get_widget("windowOcc")
        wid = self.__wtree.get_widget("frameOcc")
        self.__frame = wid.get_child()
        wid.remove(self.__frame)

        # information
        happen_button = self.__wtree.get_widget("possible")
        happen_button.set_active(True)
        self.property_happen = True
        happen_button.connect("toggled", self.rb_callback, "p")
        never_happen_button = self.__wtree.get_widget("impossible")
        never_happen_button.connect("toggled", self.rb_callback, "i")

        self.step_entry = self.__wtree.get_widget("max_step")
        self.step_entry.set_text("10")
        self.step_entry.connect("changed", self.on_entry_changed)
        self.property_entry = self.__wtree.get_widget("property_entry")
        self.property_entry.connect("changed", self.on_entry_changed)
        self.inv_prop_entry = self.__wtree.get_widget("property_entry2")
        self.inv_prop_entry.connect("changed", self.on_entry_changed)
        self.start_prop_entry = self.__wtree.get_widget("property_entry3")
        self.start_prop_entry.connect("changed", self.on_entry_changed)

        # yes/no button
        is_satisfiable_button = self.__wtree.get_widget("button_yn")
        is_satisfiable_button.connect("clicked", self.on_yn)
        # conditions button
        search_solutions_button = self.__wtree.get_widget("button_cond")
        search_solutions_button.connect("clicked", self.on_cond)

        # condition exploration
        self.max_sol_entry = self.__wtree.get_widget("entry_nbsol")
        self.max_sol_entry.set_text("10")
        self.max_sol_entry.connect("changed", self.on_entry_changed)
        self.but_solutions = self.__wtree.get_widget("but_solutions")
        self.but_solutions.connect("clicked", self.on_solutions)
        self.but_solutions.set_sensitive(False)

        self.but_mac = self.__wtree.get_widget("but_mac")
        self.but_mac.connect("clicked", self.on_mac)
        self.but_mac.connect("clicked", self.on_entry_changed)
        self.but_mac.set_sensitive(False)


        # Tooltips
        self.tooltips = gtk.Tooltips()
        self.tooltips.set_tip(
            self.property_entry,
            "Final property that will occur at the end of the simulation.\n"
            "The writing of this field obeys a syntax of propositional logic. "
            "Concretely, you must specify the name of the entities of interest "
            "linked by logical operators (NOT or AND). Parentheses are allowed.",
        )
        self.tooltips.set_tip(
            self.start_prop_entry,
            "Property that will be part of the initial state of the model.\n"
            "In concrete terms, some entities can be activated by this mechanism "
            "without modifying the model.",
        )
        self.tooltips.set_tip(
            self.inv_prop_entry,
            "Invariant property that will never occur during the simulation.\n"
            "The given logical formula will be enclose by a logical not() and "
            "will be checked at each step of the simulation.",
        )
        self.tooltips.set_tip(is_satisfiable_button, "Run a quick satisfiability test")
        self.tooltips.set_tip(
            search_solutions_button,
            "Search non optimal solutions.\n"
            "If non-optimal solutions are found, the search for optimal "
            "solutions (Minimal Activation Condition) will be unlocked.",
        )
        self.tooltips.set_tip(self.but_solutions, "Show non-optimal solutions")
        self.tooltips.set_tip(
            self.but_mac,
            "Search and display MACs (Minimal Activation Conditions).\n"
            "This process can be expensive depending on the model considered. "
            "Command line access is recommended for the heaviest tasks.",
        )

        self.tooltips.enable()
        self.tooltips.set_delay(cm.TOOLTIPS_DELAY)

        # children
        self.__aux_win = []

    def get_frame(self):
        """
        as name says
        """
        return self.__frame

    def rb_callback(self, widget, idw):
        """
        happen and never happen call back
        """
        if idw == "p":
            self.property_happen = True
        else:
            self.property_happen = False

    def on_entry_changed(self, widget):
        """
        react in cas of change in entries
        """
        #  buttons
        self.but_solutions.set_sensitive(False)
        self.but_mac.set_sensitive(False)

    def on_yn(self, widget):
        """
        lauch a satisfiability test
        """

        def is_satisfiable(query, max_step):
            """Launch the solver processing, catch/print exceptions, warn the user"""
            try:
                reachable = self.mcla.sq_is_satisfiable(query, max_step)

            except Exception as e:
                LOGGER.error("QueryCheckingForm:on_yn: %s", e)
                print(traceback.format_exc())
            finally:
                # Check the reporter
                if self.__error_reporter.error:
                    cancel_warn(self.__error_reporter.memory)
                    self.__error_reporter.reset()
                    return

            if reachable:
                ok_warn("Yes, we can!", "green")
            else:
                ok_warn("NO", "red")

        # get info
        final_prop = self.property_entry.get_text()
        if len(final_prop) == 0:
            cancel_warn("Missing property")
            return
        inv_prop = self.inv_prop_entry.get_text()
        start_prop = self.start_prop_entry.get_text()
        max_step_str = self.step_entry.get_text()
        if len(max_step_str) == 0:
            cancel_warn("Number of steps not specified")
            return
        max_step = int(max_step_str)
        self.mcla = MCLAnalyser(self.__error_reporter)
        self.mcla.build_from_chart_model(self.__model)
        if self.__error_reporter.error:
            cancel_warn(self.__error_reporter.memory)
            self.__error_reporter.reset()
            return

        # solve happen/never happen problem
        if self.property_happen:
            if len(inv_prop) == 0:
                inv_prop = None
            else:
                # Enclose the invariant property in a logical not
                inv_prop = "not (" + inv_prop + ")"

            if len(start_prop) == 0:
                start_prop = None

            query = MCLSimpleQuery(start_prop, inv_prop, final_prop)

        else:
            if len(inv_prop) != 0:
                # Enclose final AND invariant properties in a logical not
                final_prop = "not (" + final_prop + " and " + inv_prop + ")"

            query = MCLSimpleQuery(start_prop, inv_prop, None)

        is_satisfiable(query, max_step)

    def on_cond(self, widget):
        """Launch solution computations
        If solutions are found, MAC search is unlocked.
        """

        def search_solutions(query, max_step, max_sol):
            """Launch the solver processing, catch/print exceptions, warn the user"""
            try:
                solutions = self.mcla.sq_frontier_solutions(query, max_step, max_sol)

                LOGGER.debug(
                    "QueryCheckingForm.on_cond:: Solutions: %s", len(solutions)
                )

            except Exception as e:
                LOGGER.error("QueryCheckingForm:on_cond: %s", e)
                print(traceback.format_exc())
            finally:
                # Check the reporter
                if self.__error_reporter.error:
                    cancel_warn(self.__error_reporter.memory)
                    self.__error_reporter.reset()
                    return

            return solutions

        # desactivate buttons
        self.but_solutions.set_sensitive(False)
        self.but_mac.set_sensitive(False)

        # get info
        final_prop = self.property_entry.get_text()
        if len(final_prop) == 0:
            cancel_warn("Missing property")
            return
        inv_prop = self.inv_prop_entry.get_text()
        start_prop = self.start_prop_entry.get_text()
        max_step_str = self.step_entry.get_text()
        if len(max_step_str) == 0:
            cancel_warn("Number of steps not specified")
            return
        max_step = int(max_step_str)
        max_sol_str = self.max_sol_entry.get_text()
        if len(max_sol_str) == 0:
            cancel_warn("Number of solutions not specified")
            return
        max_sol = int(max_sol_str)

        self.mcla = MCLAnalyser(self.__error_reporter)
        self.mcla.build_from_chart_model(self.__model)
        if self.__error_reporter.error:
            cancel_warn(self.__error_reporter.memory)
            self.__error_reporter.reset()
            return

        # solve happen/never happen problem
        if self.property_happen:
            LOGGER.debug("QueryCheckingForm.on_cond:: Solve happen problem!")

            if len(inv_prop) == 0:
                inv_prop = None
            else:
                # Enclose the invariant property in a logical not
                inv_prop = "not (" + inv_prop + ")"

            if len(start_prop) == 0:
                start_prop = None

            query = MCLSimpleQuery(start_prop, inv_prop, final_prop)

        else:
            LOGGER.debug("QueryCheckingForm.on_cond:: Solve never happen problem!")

            if len(inv_prop) != 0:
                # Enclose final AND invariant properties in a logical not
                final_prop = "not (" + final_prop + " and " + inv_prop + ")"

            query = MCLSimpleQuery(start_prop, inv_prop, None)

        self.lsol = search_solutions(query, max_step, max_sol)

        if self.lsol is None:
            return

        # affectation
        self.property = final_prop
        self.inv_prop = inv_prop
        self.start_prop = start_prop
        self.max_step = max_step
        self.max_sol = max_sol

        if len(self.lsol) == 0:
            cancel_warn("No solution")
            return

        # activate buttons
        self.but_solutions.set_sensitive(True)
        self.but_mac.set_sensitive(True)

    def on_solutions(self, widget):
        """
        display solutions
        """
        SolutionWindow(self.lsol, self.__emvc, self.__error_reporter, self)

    def on_mac(self, widget):
        """
        launch mac computations
        """
        ask = confirm(
            None,
            "This process takes time.\n"
            + "The interface may freeze during this time.\n"
            + "Do you want to continue?",
        )
        if ask:
            query = MCLSimpleQuery(self.start_prop, self.inv_prop, self.property)
            mac_list = tuple(self.mcla.mac_search(query, self.max_step))
            if not mac_list:
                ok_warn(
                    "The solver could not find a MAC,\n"
                    + "you should refine your query."
                )
            else:
                # MACs solutions
                SolutionWindow(
                    mac_list,
                    self.__emvc,
                    self.__error_reporter,
                    self,
                    title=str(len(mac_list)) + " Minimal Activation Conditions",
                )

    # sub windows management
    def win_register(self, win):
        """
        register sub window for destroy
        """
        self.__aux_win.append(win)

    def win_remove(self, win):
        """
        used when the sub window destroys itself
        """
        if win in self.__aux_win:
            self.__aux_win.remove(win)

    def clean_subwin(self):
        """
        destroy all registered subwindows
        """
        for win in self.__aux_win:
            win.destroy()

    def destroy(self):
        """
        destroy all
        """
        self.clean_subwin()
        if self.__window:
            self.__window.destroy()


class SolutionWindow(object):
    """Show solutions (non-optimal or Minimal Activation Conditions)"""

    def __init__(self, l_fsol, emvc, reporter, parent, title=""):
        """
        @param l_fsol: list<FrontierSolution>
        @param emvc: edit_mvc
        @param reporter: Classical reporter
        @param parent: parent widget
        """
        self.emvc = emvc
        self.reporter = reporter
        # frontier place list extraction and ic sequence extraction
        self.l_solutions = []
        self.l_icseq = []
        for fsol in l_fsol:
            self.l_solutions.append(fsol.sorted_activated_frontier)
            self.l_icseq.append(fsol.ic_sequence)

        self.aux_win = []
        self.simul = None

        self.compt = 0
        self.choice = 0
        self.textview_list = []

        # window creation
        template = pkg_resources.resource_filename(
            __name__, "../chart_glade/checker_solutions.glade"
        )
        self.wtree = gtk.glade.XML(template)
        self.window = self.wtree.get_widget("window1")
        # Favicon
        favicon = pkg_resources.resource_filename(
            __name__,
            "../images/favicon.ico"
        )
        self.window.set_icon_from_file(favicon)

        self.window.set_resizable(True)
        height = gtk.gdk.screen_height()
        height = int(height * 0.30)
        self.window.set_size_request(700, height)
        # Ability to destroy opened ChartSimulControler
        self.window.connect("destroy", self.on_destroy)
        # Event on_escape key pressed
        self.window.connect("key_press_event", self.on_escape)

        # Info label: Set title
        if title:
            info = self.wtree.get_widget("label_nbsol")
            info.set_text(title)
        else:
            info = self.wtree.get_widget("label_nbsol")
            info.set_text(str(len(self.l_solutions)) + " solutions found")

        # button save choice
        save_choice_button = self.wtree.get_widget("but_sv_choice")
        save_choice_button.connect("clicked", self.on_save, False)

        # button save all
        save_all_button = self.wtree.get_widget("but_sv_all")
        save_all_button.connect("clicked", self.on_save, True)
        if len(self.l_solutions) == 1:
            save_all_button.set_sensitive(False)

        # button frame
        self.frame_sol = self.wtree.get_widget("frame_sol")
        vbox = gtk.VBox(False, 0)
        self.frame_sol.add(vbox)
        vbox.show()

        scroll = gtk.ScrolledWindow()
        vbox.pack_start(scroll)

        vbox2 = gtk.VBox()
        scroll.add_with_viewport(vbox2)
        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.button_group = None

        frame = self.build_button_frame(self.l_solutions[0])
        vbox2.pack_start(frame)

        for sol in self.l_solutions[1:]:
            hsep = gtk.HSeparator()
            vbox2.pack_start(hsep, False, False)

            frame = self.build_button_frame(sol, self.button_group)
            vbox2.pack_start(frame)

        # Buttons
        # button inhibitor => Disabled unknown function, See .glade to reactivate
        # inhibitor_button = self.wtree.get_widget("but_inhib")
        # inhibitor_button.connect("clicked", self.on_inhib)
        # inhibitor_button.set_sensitive(False)
        # button simulation
        simulation_button = self.wtree.get_widget("but_simul")
        simulation_button.connect("clicked", self.on_simul)
        # button chronogram
        chronogram_button = self.wtree.get_widget("but_chrono")
        chronogram_button.connect("clicked", self.on_chrono)

        # Tooltips
        self.tooltips = gtk.Tooltips()
        self.tooltips.set_tip(
            save_choice_button,
            "Export the selected solution and its trajectory of events to a file",
        )
        self.tooltips.set_tip(
            save_all_button,
            "Export all the solutions and their trajectories of events to a file",
        )
        self.tooltips.set_tip(
            simulation_button,
            "Replay the selected trajectory step by step on the model",
        )
        self.tooltips.set_tip(
            chronogram_button,
            "Display the chronogram of the activation of the places in "
            "the selected trajectory",
        )

        self.tooltips.enable()
        self.tooltips.set_delay(cm.TOOLTIPS_DELAY)

        # register
        parent.win_register(self)
        self.parent = parent

        # display
        self.window.show_all()
        # gtk.main()

    def on_destroy(self, widget):
        """
        when leaving the window
        """
        self.clean_subwin()
        if self.simul:
            self.simul.on_destroy(None)
        self.win_remove(self)

    def destroy(self):
        """
        when leaving the window
        """
        self.clean_subwin()
        if self.simul:
            self.simul.on_destroy(None)
        if self.window:
            self.window.destroy()

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.window.destroy()

    def build_button_frame(self, state_list, group=None):
        """
        ???
        """
        template = pkg_resources.resource_filename(
            __name__, "../chart_glade/button_frame.glade"
        )
        wtree = gtk.glade.XML(template)
        wid = wtree.get_widget("button_frame")
        frame = wid.get_child()
        wid.remove(frame)

        # textview
        tname = "tw" + str(self.compt)
        text = wtree.get_widget("textview")
        text.set_name(tname)

        text_buffer = text.get_buffer()
        str_state = "\n"
        for state in state_list:
            str_state += state + "\t"
        str_state += "\n"
        text_buffer.set_text(str_state)
        text.set_editable(False)

        self.textview_list.append(state_list)

        # radio button
        rname = "rb" + str(self.compt)
        rbut = wtree.get_widget("rButton")
        rbut.set_name(rname)

        if group:
            rbut.set_group(group)
        else:
            self.button_group = rbut
            rbut.set_active(True)

        rbut.connect("toggled", self.rb_callback, self.compt)
        self.compt += 1
        return frame

    def rb_callback(self, widget, id_frame):
        """
        radio button call_back
        """
        self.choice = id_frame

    def on_save(self, widget, all_sol=True):
        """
        open a window to save as xml file
        @param all_sol : boolean - True if we take all solutions
        """
        choice = gtk.FileChooserDialog(
            "Save solution",
            None,
            gtk.FILE_CHOOSER_ACTION_SAVE,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE, gtk.RESPONSE_OK),
        )
        choice.set_default_response(gtk.RESPONSE_OK)

        # add a filter to see only xml files for biochart (*.txt)
        ffilter = gtk.FileFilter()
        ffilter.set_name("txt files")
        ffilter.add_pattern("*.txt")
        choice.add_filter(ffilter)

        # add a filter to see all
        no_filter = gtk.FileFilter()
        no_filter.set_name("all")
        no_filter.add_pattern("*")
        choice.add_filter(no_filter)

        response = choice.run()
        if response == gtk.RESPONSE_OK:
            self.create_solution_file(choice.get_filename(), all_sol)
        elif response == gtk.RESPONSE_CANCEL:
            pass
            # print 'Closed, no files selected'
        choice.destroy()

    def create_solution_file(self, sol_file, all_bool):
        """
        make a xml file with the current model
        """
        if all_bool:
            init_places = self.l_solutions
            input_clocks = self.l_icseq
        else:
            init_places = [self.l_solutions[self.choice]]
            input_clocks = [self.l_icseq[self.choice]]

        sfile = open(sol_file, "w")
        for i in range(len(init_places)):
            spla = init_places[i]
            icseq = input_clocks[i]
            for elemt in spla:
                sfile.write(str(elemt) + "\t")
            for ica in icseq:
                sfile.write("\n" + ica)
            sfile.write("\n")
        sfile.close()

    def on_inhib(self, widget):
        """
        Compute inhibitors
        """
#
#        start_list = self.l_solutions[self.choice]
#        start_p = logical_and(start_list)
#        inv_p = 'not('+self.parent.property+')'
#        mac_list = self.parent.mcla.mac_search(start_p, inv_p, None,
#                                               self.parent.max_step)
#        inhib_list = []
#        for mac in mac_list:
#            im = []
#            for var in mac:
#                if var in start_list:
#                    continue
#                else:
#                    im.append(var)
#            if im:
#                inhib_list.append(im)
#        if len(inhib_list) == 0:
#            cancel_warn("This solution has not inhibitor")
#        else :
#            Inhib_Window(inhib_list, self.emvc, self.reporter, self)
        pass

    def on_simul(self, widget):
        """Launch simulation of a solution

        .. note:: We also parse the start property since it can activate some
            places (other than frontiers), whose are required for the simulator
            to operate properly.
            Indeed, the simulator accepts to fire transitions only if places
            are activated for a given step.

        .. note:: Mentionner une place frontière en start property va l'activer
            directement et graphiquement; init_places va donc la contenir et
            les solutions également.
            **Cela n'est pas le cas pour une place autre, non frontière**.
            Ici nous rétablissons simplement la liste de places init_places pour
            que le simulateur fonctionne correctement en activant dans le
            graph editor les bonnes places/transitions à chaque étape.

        .. todo:: Mettre à jour les solutions en y ajoutant les places dans
            start property?
        """
        if self.simul:
            self.simul.on_destroy(None)

        # Cast to avoid further modification of the original data
        init_places = list(self.l_solutions[self.choice])

        start_property = self.parent.start_prop
        self.emvc.model.transition_list

        if start_property:
            # Start property parsing
            LOGGER.debug(
                "SolutionWindow:on_simul:\n"
                "   start_property: %s\n"
                "   init_places: %s\n",
                start_property,
                init_places,
            )

            tree_prop = compile_cond(
                start_property,
                # Fake symbol_table: for the parser sigexpr_compiler
                # {'A': ('state', 0), 'C': ('state', 0), ...}
                defaultdict(lambda: ("state", 0)),
                CompilReporter(),
            )

            LOGGER.debug(
                "SolutionWindow:on_simul: tree prop %s %s", tree_prop, repr(tree_prop)
            )

            prop_visitor = PropertyVisitor()
            tree_prop.accept(prop_visitor)
            init_places += prop_visitor.nodes

            LOGGER.debug("SolutionWindow:on_simul: init_places", init_places)


        input_clocks = self.l_icseq[self.choice]
        LOGGER.debug("SolutionWindow:on_simul: clocks", input_clocks)

        # Remove duplicates
        init_places = list(set(init_places))

        self.simul = ChartSimulControler(
            self.emvc, True, self.reporter, init_places, input_clocks, self
        )

    def on_chrono(self, widget):
        """
        Show a chronogram of a trajectory initialized with solution
        """
        # stand alone simulator
        sim = ChartSimulator()
        sim.build(self.emvc.model, True, self.reporter)

        chrono = ChartChrono.for_trajectory(sim)
        self.win_register(chrono)

        # simulation
        init_places = self.l_solutions[self.choice]
        input_clocks = self.l_icseq[self.choice]
        sim.simul_init_places(init_places)
        # input buffer is any input
        if input_clocks:
            sim.set_act_input_stream(input_clocks)
        i = 0
        while i <= self.parent.max_step:
            chrono.register(sim)
            try:
                sim.simul_step()
                i += 1
            except Exception:
                break
        # clean model
        self.emvc.model.clean()
        self.emvc.model.notify()
        # display
        chrono.display_selected_chrono()

    # sub windows management
    def win_register(self, win):
        """
        sub windows management
        """
        if win.__class__ == ChartSimulControler:
            self.simul = win
        self.aux_win.append(win)

    def win_remove(self, win):
        """
        sub windows management
        """
        if win in self.aux_win:
            self.aux_win.remove(win)
        if win == self.simul:
            self.simul = None

    def clean_subwin(self):
        """
        sub windows management
        """
        for win in self.aux_win:
            win.destroy()
        self.aux_win = []


class PropertyVisitor(object):
    """Translate a SigExpression (condition of a transition) into a set of nodes
    that satisfy the condition.

    Used to force the simulator to use nodes mentioned in a start property by
    the user.

    Non mandatory nodes (inhibitors) are not retained.
    Activators can be accessed in the attribute 'nodes'
    """

    def __init__(self):
        self.nodes = list()
        # Flag used to detect if we are in a leaf on in the root of the tree
        # If this is the root, we can set the attribute self.nodes and return
        self.root = True

    def visit_sig_ident(self, tnode):
        """ident -> keep the node"""
        # print("ident node", str(tnode))

        node_name = [tnode.name]
        if self.root:
            self.nodes = node_name
        else:
            return node_name

    def visit_sig_not(self, node):
        """translate a not

        Do not retain the subnode if it is an "ident" node.
        Retain only subnodes if they are in a binary expression linked by a logical "and".

        Because:

            - not (a or b) is valid only if a and b are False;
              thus we don't care of a and b
            - not (a and b) is valid only if a or b are True, or if a and b are False;
              thus a and b can be True alternately and it is sufficient for us.
        """
        # print("not node:", str(node))
        root = self.root
        self.root = False
        subnodes = node.operand.accept(self)

        nodes = []

        if isinstance(node, SigBinExpr):
            # print("operator:", node.operator)

            if node.operator == "and":
                nodes = subnodes

        if root:
            self.nodes = nodes
            return
        return []

    def visit_sig_sync(self, binnode):
        """translate a binary (or/and) expression"""
        # recursive visits
        root = self.root
        self.root = False

        # print("sig sync found", str(binnode))
        nl1 = binnode.left_h.accept(self)
        nl2 = binnode.right_h.accept(self)
        # print("sig sync end", binnode.operator, nl1, nl2, str(binnode))

        nodes = nl1 + nl2

        if root:
            self.nodes = nodes
            return
        return nodes


class InhibWindow(SolutionWindow):
    """Show inhibitors of the current solution

    Not used anymore (code disabled)
    """

    def __init__(self, l_solutions, emvc, reporter, parent):

        SolutionWindow.__init__(
            self,
            l_solutions,
            emvc,
            reporter,
            parent,
            title=str(len(l_solutions)) + " inhibitors",
        )

        # button inhibitor
        button = self.wtree.get_widget("but_inhib")
        button.connect("clicked", self.on_inhib)
        button.set_sensitive(False)


def logical_and(list):
    """
    @return: logical_formula: str - AND of the input list
    """

    if len(list) == 0:
        return
    elif len(list) == 1:
        return list[0]
    else:
        logical_formula = ""
        for elemnt in list:
            logical_formula += elemnt + " and "
        logical_formula = logical_formula[:-5]
        #        print logical_formula
        return "(" + logical_formula + ")"



## Filename    : layout
## Author(s)   : Michel Le Borgne
## Created     : 01/2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA/IRSET
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s):
##
"""
Layouts are performed by a visitor of the chart model
"""
import pygraphviz as pgv

import cadbiom.commons as cm

LOGGER = cm.logger()


class LayoutVisitor(object):
    """Visitor used to do a gt graph layout

    Available layouts::

        - dot - "hierarchical" or layered drawings of directed graphs.
        This is the default tool to use if edges have directionality.

        - neato - "spring model" layouts.
        This is the default tool to use if the graph is not too large
        (about 100 nodes) and you don't know anything else about it.
        Neato attempts to minimize a global energy function, which is equivalent
        to statistical multi-dimensional scaling.

        - fdp - "spring model" layouts similar to those of neato, but does
        this by reducing forces rather than working with energy.

        - sfdp - multiscale version of fdp for the layout of large graphs.

        - twopi - radial layouts, after Graham Wills 97.
        Nodes are placed on concentric circles depending their distance from a
        given root node.

        - circo - circular layout, after Six and Tollis 99, Kauffman and Wiese 02.
        This is suitable for certain diagrams of multiple cyclic structures,
        such as certain telecommunications networks.

    .. seealso:: https://www.graphviz.org/

    """

    def __init__(self, view, layout_style):
        self.view = view
        self.drawing_style = view.drawing_style
        self.layout = layout_style

    def visit_chart_model(self, model):
        """
         Not used
        """
        model.get_root().accept(self)

    def visit_cstart_node(self, node):
        """
         Not used
        """
        return

    def visit_ctrap_node(self, node):
        """
         Not used
        """
        return

    def visit_csimple_node(self, node):
        """
         Not used
        """
        return

    def visit_cinput_node(self, node):
        """
         Not used
        """
        return

    def visit_cperm_node(self, node):
        """
         Not used
        """
        return

    def visit_cmacro_node(self, node):
        """
        Layouts are done on each macro place
        Change the local coordinates of subnodes according to layout style
        """
        if node.sub_nodes == []:
            return
        agraph = pgv.AGraph(splines="ortho")
        node_dict = dict()
        # build nodes
        for snode in node.sub_nodes:
            agraph.add_node(snode.name)
            anode = agraph.get_node(snode.name)
            node_dict[snode.name] = snode
            if snode.is_macro() and snode.model.show_macro:
                anode.attr["width"] = "%s" % snode.wloc
                anode.attr["height"] = "%s" % snode.hloc
                snode.accept(self)
        # build edges
        for tgr in node.transitions:
            for trans in tgr:
                agraph.add_edge(trans.ori.name, trans.ext.name)
        # layout
        layout_style = self.layout
        if layout_style == "hierarchical_LR":
            agraph.layout(prog="dot", args="-Grankdir=LR")
        elif layout_style == "hierarchical_TB":
            agraph.layout(prog="dot", args="-Grankdir=BT")
        elif layout_style == "neato":
            agraph.layout(prog="neato")
        elif layout_style == "fdp":
            agraph.layout(prog="fdp")
        elif layout_style == "sfdp":
            agraph.layout(prog="sfdp")
        elif layout_style == "twopi":
            agraph.layout(prog="twopi")
        elif layout_style == "circo":
            agraph.layout(prog="circo")
        else:
            LOGGER.error(
                "LayoutVisitor::visit_cmacro_node: Unknow layout: %s", layout_style
            )
            exit(1)

        # change coordinates
        # bounding box for ratios
        bbox = pgv.graphviz.agget(agraph.handle, "bb")
        bbox = bbox.split(",")
        xmin = float(bbox[0])
        xmax = float(bbox[2])
        ymin = float(bbox[1])
        ymax = float(bbox[3])
        # adjust graphic window
        view = self.view
        cond = (xmax - xmin) / 3.0 > view.draw_width
        cond = cond or ((xmax - xmin) / 3.0 < view.draw_width / 2.0)
        if cond:
            view.draw_width = min(int((xmax - xmin) / 3.0), 10400)  # 31200 max
            view.draw_width = max(1300, view.draw_width)
            view.min_width = view.draw_width
        cond = ymax - ymin > view.draw_height
        cond = cond or ymax - ymin < view.draw_height / 2.0
        if cond:
            view.draw_height = min(int(ymax - ymin), 7200)  # 21600 max
            view.draw_height = max(900, view.draw_height)
            view.min_height = view.draw_height
        # reset zoom counter
        view.zoom_count = 0
        gwidth = 1.0
        gheight = 1.0  # virtual window for the macro node
        if xmin == xmax or ymin == ymax:
            LOGGER.error(
                "LayoutVisitor::visit_cmacro_node: Coordinates consistency: "
                "x:(%s vs %s); y:(%s vs %s)",
                xmin, xmax,
                ymin, ymax,
            )
            exit(1)

        x_ratio = gwidth / (xmax - xmin)
        y_ratio = gheight / (ymax - ymin)
        # changing coordinates
        for node in agraph.nodes():
            pos = node.attr["pos"].split(",")
            cnn = node_dict[node]
            xcoord = float(pos[0]) * x_ratio
            ycoord = float(pos[1]) * y_ratio
            cnn.set_layout_coordinates(xcoord, ycoord)

    def visit_ctop_node(self, node):
        """
        Same as macro node
        """
        self.visit_cmacro_node(node)
        return

    def visit_ctransition(self, trans):
        """
         Not used
        """
        return



## Filename    : charter.py
## Author(s)   : Geoffroy Andrieux
## Created     : 03/2010
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne
##
"""
Small widgets for Cadbiom gui::

    - SearchManager: List of simple nodes for searching
    - SearchFrontier: Same as SearchManager but specialized on frontier nodes
    - LegendWindow: Widget to display the legend
    - ImportPIDParam: Widget for model importing
    - ImportBioPAXParams: Widget for BioPAX model importing
"""
# Standard imports
import itertools as it
import pkg_resources
import re
import traceback

# Custom imports
import gtk

# Cadbiom imports
from utils.listDisplay import ToggleList
from utils.fileHandling import FileChooser
from utils.reporter import CompilReporter
from utils.warn import cancel_warn, confirm
from cadbiom.models.guard_transitions.analyser.ana_visitors import FrontierVisitor
from cadbiom.models.guard_transitions.chart_model import ChartModel, CStartNode
from cadbiom.models.guard_transitions.analyser import model_corrections
from cadbiom.models.guard_transitions.translators.chart_xml import MakeModelFromXmlFile
from cadbiom.models.guard_transitions.translators.chart_xml_pid import (
    MakeModelFromPidFile,
)
from chart_simulator.chart_simul_controler import DisplayError
from cadbiom import commons as cm

LOGGER = cm.logger()


class SearchManager(object):
    """List of simple nodes for searching


    :param model_changed: Flag updated to True by the model.
        Its value is True until the user clicks on the "update" button.
        This operation is manual since it can be costly.
    :type model_changed: <boolean>
    """

    def __init__(self, chart, notebook, label):
        self.charter = chart
        self.model = None
        self.model_changed = True
        self.notebook = notebook

        # simple nodes
        self.sn_frame = gtk.Frame()
        vbox = gtk.VBox(False, 0)
        self.sn_frame.add(vbox)
        # simple nodes list
        self.sn_viewer = ToggleList()
        label = gtk.Label(label)
        # wrap into scrollwindow
        scroll = gtk.ScrolledWindow()
        scroll.add_with_viewport(self.sn_viewer)
        vbox.pack_start(scroll, True, True, 0)
        # search entry
        self.search_entry = gtk.Entry()
        vbox.pack_start(self.search_entry, False, False, 0)
        # buttons
        hbox = gtk.HBox()
        vbox.pack_start(hbox, False, False, 0)
        update_button = gtk.Button("Update")
        update_button.connect("clicked", self.on_update)
        hbox.pack_start(update_button, False, False, 0)
        show_button = gtk.Button("Show")
        show_button.connect("clicked", self.on_show)
        hbox.pack_start(show_button, False, False, 0)
        extract_button = gtk.Button("Extract")
        extract_button.connect("clicked", self.on_extract)
        hbox.pack_start(extract_button, False, False, 0)
        clear_button = gtk.Button("Clear")
        clear_button.connect("clicked", self.on_clear)
        hbox.pack_start(clear_button, False, False, 0)
        notebook.append_page(self.sn_frame, label)


        # Tooltips
        self.tooltips = gtk.Tooltips()
        self.tooltips.set_tip(update_button, "Sync the list of nodes with the model")
        self.tooltips.set_tip(show_button, "Show the selected nodes on the model")
        self.tooltips.set_tip(extract_button, "Extract the selected nodes to another model")
        self.tooltips.set_tip(clear_button, "Reset the selection")
        self.tooltips.set_tip(
            self.search_entry,
            "Search for nodes by their names.\nYou can use regular expressions"
        )

        self.tooltips.enable()
        self.tooltips.set_delay(cm.TOOLTIPS_DELAY)

    def update(self):
        """The model informs observers that the model has changed

        We keep this value to True until the user clicks on the "update" button.
        """
        self.model_changed = True

    def display_nodes(self):
        """Display node names on the widget

        - On :meth:`set_model` at the initialization of the model
        - On :meth:`on_update` by a click on the button "update" by the user
        """
        # get node names
        lnode = self.model.get_simple_node_names()
        lnode.sort()
        # display node names
        self.sn_viewer.refresh(lnode)

    def set_model(self, model):
        """Refresh the GUI with the given model

        Called by :meth:`~cadbiom_gui.gt_gui.edit_mvc.set_current_edit_mvc`
        during the switch to a new/another tab/model.
        """
        if self.model:
            self.model.detach(self)
        self.model = model
        self.model.attach(self)
        self.model_changed = False
        # display node names
        self.display_nodes()
        # Check/tick previously selected nodes
        self.sn_viewer.set_selected_items(
            [node.name for node in self.model.marked_nodes]
        )

    def on_clear(self, widget):
        """Clear button callback: Reset selected nodes, and search entry field
        """
        self.sn_viewer.deselect_all()
        self.search_entry.set_text("")
        if self.model:
            self.model.unset_search_mark()
            self.model_changed = False  # search_mark implies a notify

    def on_update(self, widget):
        """Update button callback: Sync the view of simple nodes with the model"""
        if not self.model:
            return
        # unmarking nodes and transitions with conditions
        self.model.unset_search_mark()
        # display node names
        self.display_nodes()
        self.model_changed = False

    def on_show(self, widget):
        """Show button callback: Show nodes of interest in the graph editor

        Transitions with conditions where the selected nodes are involved
        are also marked
        """
        if not self.model:
            return
        # check model
        if self.model_changed:
            action = confirm(
                None,
                "The model has been modified since the last synchronization.\n"
                "Would you like to continue without updating the set of nodes "
                "and transitions?"
            )
            if not action:
                return

        # Get marked nodes names
        node_names = self.get_selected_or_matching_node_names()
        if not node_names:
            return

        # Set the mark status of all selected nodes/transitions
        # Transitions with conditions where the selected nodes are involved
        # are also marked
        self.model.set_search_mark(node_names)

        # Check/Display the marks in the nodes list
        self.sn_viewer.set_selected_items(node_names)
        self.model_changed = False  # search_mark implies a notify

    def on_extract(self, widget):
        """Extract nodes environment

        - Get selected nodes
        - Get all transitions of interest:
          - ori and ext nodes are concerned
          - ori or ext are concerned and the second node is a Start/Trap node
        """
        if not self.model:
            return
        # check model
        if self.model_changed:
            action = confirm(
                None,
                "The model has been modified since the last synchronization.\n"
                "Would you like to continue without updating the set of nodes "
                "and transitions before the extraction?"
            )
            if not action:
                return

        # Get marked nodes names
        node_names = self.get_selected_or_matching_node_names()
        if not node_names:
            return

        # Init new model
        submodel = ChartModel(self.model.name + "_extract")
        tnode = submodel.get_root()

        # Convert nodes names to nodes objects
        old_nodes = {self.model.node_dict[name] for name in node_names}

        # Get all transitions of interest:
        # - ori and ext nodes are concerned
        # - ori or ext are concerned and the second node is a Start/Trap node
        old_transitions = set()
        for transition in self.model.transition_list + list(
            it.chain(*self.model.signal_transition_dict.values())
        ):
            if {transition.ori, transition.ext}.issubset(old_nodes):
                pass
            elif transition.ori in old_nodes and isinstance(transition.ext, CStartNode):
                old_nodes.add(transition.ext)
            elif transition.ext in old_nodes and isinstance(transition.ori, CStartNode):
                old_nodes.add(transition.ori)
            else:
                continue

            # Keep the transition
            old_transitions.add(transition)

        # LOGGER.debug("Extract old nodes: %s", old_nodes)
        # LOGGER.debug("Extract old transitions: %s", old_transitions)
        # Some nodes are involved in the conditions of the transitions of interest
        # => These nodes have to be extracted
        old_nodes.update([
            self.model.node_dict[node_name]
            for node_name, transitions in self.model.signal_transition_dict.items()
            # Transition of interest encountered ?
            if set(transitions) & old_transitions
        ])

        # LOGGER.debug("Extract old nodes with extra cond nodes: %s", old_nodes)

        ## New model construction
        # Add nodes
        nodes_mapping = {old_node: tnode.add_copy(old_node) for old_node in old_nodes}
        # Add transitions
        for transition in old_transitions:
            new_trans = tnode.add_transition(
                nodes_mapping[transition.ori],
                nodes_mapping[transition.ext]
            )
            new_trans.set_condition(transition.condition)
            new_trans.set_event(transition.event)
            new_trans.set_action(transition.action)
            new_trans.set_note(transition.note)

        # Display in charter
        self.charter.add_edit_mvc(submodel.name, submodel, False)
        self.charter.do_layout(None, "hierarchical_LR")

    def get_selected_or_matching_node_names(self):
        """Get nodes names of interest

        Methods used to find nodes:
            - nodes selected in the list of the GUI
            - parse regex inserted in the search entry

        .. note:: selected items work in conjunction with regular expressions
            in the search entry.

        :return: List of names or None
        :rtype: <list <str>> or None
        """
        node_names = self.sn_viewer.get_selected_items()

        # If no nodes selected, use search entry
        regex = self.search_entry.get_text()
        if not regex:
            return node_names

        try:
            regex_obj = re.compile(regex)
        except Exception as e:
            cancel_warn("Incorrect regular expression:\n {}".format(e))
            return

        return list(set(node_names + self.model.get_matching_node_names(regex_obj)))

    def get_selected_items(self):
        """
        As it says
        """
        return self.sn_viewer.get_selected_items()

    def clear(self):
        """Remove all items"""
        self.sn_viewer.clear()


class SearchFrontier(SearchManager):
    """Same as SearchManager but specialized on frontier nodes
    """

    def __init__(self, chart, notebook, label):
        SearchManager.__init__(self, chart, notebook, label)

    def display_nodes(self):
        """Display node names on the widget"""
        # get node names
        lnode = self.get_frontier_node_names()
        lnode.sort()
        # display node names
        self.sn_viewer.refresh(lnode)

    def get_frontier_node_names(self):
        """
        As it says
        """
        fvi = FrontierVisitor()
        self.model.accept(fvi)
        return fvi.frontier


class LegendWindow(object):
    """
    Widget to display the legend
    """

    def __init__(self, parent=None):
        self.win = gtk.Window()
        self.win.set_title("CADBIOM-Chart Legend")
        self.win.set_position(gtk.WIN_POS_CENTER)
        self.win.connect("destroy", self.on_destroy)
        self.win.connect("key_press_event", self.on_escape)
        # Favicon
        favicon = pkg_resources.resource_filename(
            __name__,
            "images/favicon.ico"
        )
        self.win.set_icon_from_file(favicon)

        # Legend
        image = gtk.Image()
        template = pkg_resources.resource_filename(
            __name__,
            "images/legend.png"
        )
        image.set_from_file(template)
        image.show()

        self.win.add(image)
        self.win.show_all()
        # register itself for parent or emvc
        if parent:
            parent.win_register(self)

        self.parent = parent

    def on_destroy(self, widget):
        """
        standard destroy callback
        """
        if self.parent:
            self.parent.win_remove(self)

    def destroy(self):
        """
        if registered as a child
        """
        if self.win:
            self.win.destroy()

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.destroy()


class ImportPIDParam(object):
    """
    Widget for model importing
    """

    def __init__(self, chart, parent=None):
        self.charter = chart

        # window creation
        template = pkg_resources.resource_filename(
            __name__,
            "chart_glade/import_parameter.glade"
        )
        self.wtree = gtk.glade.XML(template)
        self.main_window = self.wtree.get_widget("window1")
        self.main_window.set_title("Import window")

        self.main_window.set_resizable(True)
        hei = gtk.gdk.screen_height()
        hei = int(hei * 0.20)
        self.main_window.set_size_request(350, hei)

        # Set modal mode for the window (above all windows & block inputs)
        self.main_window.set_modal(True)
        if parent:
            self.main_window.set_transient_for(parent)

        if self.main_window:
            self.main_window.connect("destroy", self.on_destroy)
            self.main_window.connect("key_press_event", self.on_escape)

        self.main_window.set_position(gtk.WIN_POS_CENTER)

        # init param
        self.ai_inter = 0
        self.has_clock = False

        # interpretation radio button
        rbut = self.wtree.get_widget("or_rb")
        rb_name = rbut.get_name()
        rbut.connect("toggled", self.ai_inter_rb_callback, rb_name)

        rbut = self.wtree.get_widget("and_rb")
        rb_name = rbut.get_name()
        rbut.connect("toggled", self.ai_inter_rb_callback, rb_name)

        # Uncomment to add new buttons
#        rb = self.wtree.get_widget("or_and_rb")
#        rb_name = rb.get_name()
#        rb.connect("toggled", self.ai_inter_rb_callback, rb_name)
#
#        rb = self.wtree.get_widget("and_or_rb")
#        rb_name = rb.get_name()
#        rb.connect("toggled", self.ai_inter_rb_callback, rb_name)

        # clock radio button
        rbut = self.wtree.get_widget("withoutClock_rb")
        rbut.connect("toggled", self.clock_rb_callback, False)

        rbut = self.wtree.get_widget("withClock_rb")
        rbut.connect("toggled", self.clock_rb_callback, True)

        # import button
        button = self.wtree.get_widget("importButton")
        button.connect("clicked", self.on_import)

        # display
        self.main_window.show_all()

    def ai_inter_rb_callback(self, widget, ai_name):
        """
        set activator/inhibitor interpretation (and or or)
        """
        if ai_name == "or_rb":
            self.ai_inter = 0
        elif ai_name == "and_rb":
            self.ai_inter = 1

        # Uncomment to add new buttons
#        elif ai_name == "or_and_rb":
#            self.ai_inter = 2
#        else :
#            self.ai_inter = 3

    def clock_rb_callback(self, widget, has_clock):
        """
        set clock generation
        """
        if has_clock:
            self.has_clock = True
        else:
            self.has_clock = False

    def on_import(self, widget):
        """
        lauch import
        """
        fch = FileChooser(
            "Import from PID xml", "xml files", "*.xml", save_window=False
        )
        fch.do_action(self.import_from_pid_file)

    def import_from_pid_file(self, file):
        """
        compile a pid file
        """
        crep = CompilReporter()
        parser = MakeModelFromPidFile(file, crep, self.has_clock, self.ai_inter)
        if crep.error:
            DisplayError(crep, parser.model.name)
        else:
            model = parser.model
            model.modified = False
            self.charter.add_edit_mvc(model.name, model)
            self.destroy()

    def on_destroy(self, widget):
        """
        standard destroy callback
        """
        if self.main_window:
            self.main_window.destroy()

    def destroy(self):
        """
        for parents
        """
        self.on_destroy(None)

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.destroy()


class ImportBioPAXParams(object):
    """
    Widget for BioPAX model importing

    .. note:: requires biopax2cadbiom module
    """

    def __init__(self, chart, parent=None):
        self.charter = chart

        # biopax2cadbiom detection
        try:
            from biopax2cadbiom.commons import SPARQL_PATH
            from biopax2cadbiom.commons import SPARQL_LIMIT
        except ImportError:
            d_err = DisplayError(
                None,
                "biopax2cadbiom module is missing.",
                "biopax2cadbiom project is required to perform this task.\n"
                + "Please install it via the usual command:\n"
                + "pip install biopax2cadbiom",
            )
            d_err.main_window.set_modal(True)
            d_err.main_window.set_transient_for(parent)
            return

        # Init class variables (settings of the window)
        self.cadbiom_file = ""
        self.convert_full_graph = True
        self.graph_uris = list()
        self.provenance_uri = None
        self.numeric_compartments_names = False
        self.blacklist_file = None
        self.triplestore_url = SPARQL_PATH
        self.limit_sparql_results = SPARQL_LIMIT
        self.remove_strongly_connected_components = True

        # window creation
        template = pkg_resources.resource_filename(
            __name__,
            "chart_glade/import_BioPAX_parameters.glade"
        )
        self.wtree = gtk.glade.XML(template)
        self.main_window = self.wtree.get_widget("window1")
        self.main_window.set_title("Import window")
        self.main_window.set_resizable(True)
        self.main_window.set_position(gtk.WIN_POS_CENTER)

        # Set modal mode for the window (above all windows & block inputs)
        self.main_window.set_modal(True)
        if parent:
            self.main_window.set_transient_for(parent)

        if self.main_window:
            self.main_window.connect("destroy", self.on_destroy)
            self.main_window.connect("key_press_event", self.on_escape)

        # Init interface
        # Triplestore
        # Graphs URIs
        self.wtree.get_widget("text_graphs_uris").set_text(
            "http://reactome.org/mycobacterium"
        )
        # ProvenanceUri
        self.wtree.get_widget("text_provenance_uri").set_text("")
        # Triplestore URL
        self.wtree.get_widget("text_triplestore_url").set_text(self.triplestore_url)
        # BioPAX level
        self.wtree.get_widget("radio_biopax_level3").set_active(True)
        # BioPAX options
        self.wtree.get_widget("checkbox_convert_full_graph").set_active(False)
        self.wtree.get_widget("checkbox_full_compartment_names").set_active(True)
        self.wtree.get_widget("checkbox_remove_scc").set_active(True)

        # Connect buttons
        # Files
        self.wtree.get_widget("button_blacklist").connect(
            "clicked", self.on_import_blacklist_file
        )
        self.wtree.get_widget("button_output").connect(
            "clicked", self.on_set_output_file
        )
        self.wtree.get_widget("button_make_model").connect(
            "clicked", self.import_BioPAX_data
        )

        # Display
        self.main_window.show_all()

    def on_import_blacklist_file(self, widget):
        """
        Choose blacklist file of entities
        """
        fch = FileChooser(
            "Choose the blacklist file", "txt files", "*.txt", save_window=False
        )
        fch.do_action(self.set_blacklist_file)

    def on_set_output_file(self, widget):
        """
        Choose output model file
        """
        fch = FileChooser("Choose the output file name", "bcx files", "*.bcx")
        fch.do_action(self.set_output_file)
        # Change color to green
        self.wtree.get_widget("button_output").modify_bg(
            gtk.STATE_NORMAL, gtk.gdk.color_parse("green")
        )

    def set_blacklist_file(self, file):
        self.blacklist_file = file

    def set_output_file(self, file):
        self.cadbiom_file = file

    def import_BioPAX_data(self, widget):
        """
        Import from BioPAX data on triplestore
        """
        from biopax2cadbiom import biopax_converter

        # Return if output file is not set
        if self.cadbiom_file == "":
            red = gtk.gdk.color_parse("red")
            self.wtree.get_widget("button_output").modify_bg(gtk.STATE_NORMAL, red)
            return

        # Get parameters
        # List of graphs URIs
        self.graph_uris = (
            self.wtree.get_widget("text_graphs_uris").get_text().split(",")
        )

        # Provenance Uri
        provenance_uri = self.wtree.get_widget("text_provenance_uri").get_text()
        self.provenance_uri = provenance_uri if provenance_uri else None

        # Triplestore URL
        self.triplestore_url = self.wtree.get_widget("text_triplestore_url").get_text()

        # BioPAX level
        biopax2 = self.wtree.get_widget("radio_biopax_level2").get_active()
        self.graph_uris.append(
            "http://biopax.org/lvl2" if biopax2 else "http://biopax.org/lvl3"
        )

        # BioPAX convertion
        self.convert_full_graph = self.wtree.get_widget(
            "checkbox_convert_full_graph"
        ).get_active()
        self.numeric_compartments_names = self.wtree.get_widget(
            "checkbox_full_compartment_names"
        ).get_active()
        self.remove_strongly_connected_components = self.wtree.get_widget(
            "checkbox_remove_scc"
        ).get_active()

        # Color the button in green when the form is valid
        green = gtk.gdk.color_parse("green")
        self.wtree.get_widget("button_make_model").modify_bg(gtk.STATE_NORMAL, green)
        self.wtree.get_widget("button_make_model").set_label(
            "Make model. Please wait..."
        )

        # Do the magick
        # Build parameters for biopax2cadbiom
        params = {
            "cadbiomFile": self.cadbiom_file,
            "convertFullGraph": self.convert_full_graph,
            "listOfGraphUri": self.graph_uris,
            "pickleBackup": False,
            "pickleDir": "",  # don't care, because pickleBackup = False
            "numericCompartmentsNames": self.numeric_compartments_names,
            "blacklist": self.blacklist_file,
            "triplestore": self.triplestore_url,
            "no_scc_fix": True,  # StartNodes are added here
            "limit_sparql_results": self.limit_sparql_results,
            "provenanceUri": self.provenance_uri,
        }

        try:
            # StartNodes are added here, (not in biopax2cadbiom)
            # so we can load the model in memory
            biopax_converter.main(params)

            if self.remove_strongly_connected_components:
                model = model_corrections.add_start_nodes(self.cadbiom_file)
            else:
                model = MakeModelFromXmlFile(self.cadbiom_file).model

            self.charter.add_edit_mvc(model.name, model)
            self.destroy()

        except Exception as e:

            d_err = DisplayError(
                None,
                "Biopax2cadbiom",
                "Biopax2cadbiom exception: "
                + e.__class__.__name__
                + ". Please check logs.\n"
                + traceback.format_exc(),
            )
            d_err.main_window.set_modal(True)

    def on_destroy(self, widget):
        """
        standard destroy callback
        """
        if self.main_window:
            self.main_window.destroy()

    def destroy(self):
        """
        for parents
        """
        self.on_destroy(None)

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.destroy()
